# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'startupwidget.ui'
#
# Created by: PyQt5 UI code generator 5.12.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_StartupWidget(object):
    def setupUi(self, StartupWidget):
        StartupWidget.setObjectName("StartupWidget")
        StartupWidget.resize(421, 441)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(StartupWidget)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.lblMain = QtWidgets.QLabel(StartupWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lblMain.sizePolicy().hasHeightForWidth())
        self.lblMain.setSizePolicy(sizePolicy)
        self.lblMain.setPixmap(QtGui.QPixmap(":/pic/bibolamazi.png"))
        self.lblMain.setObjectName("lblMain")
        self.horizontalLayout.addWidget(self.lblMain)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        spacerItem2 = QtWidgets.QSpacerItem(20, 5, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout_3.addItem(spacerItem2)
        self.frame = QtWidgets.QFrame(StartupWidget)
        self.frame.setObjectName("frame")
        self.gridLayout = QtWidgets.QGridLayout(self.frame)
        self.gridLayout.setHorizontalSpacing(5)
        self.gridLayout.setObjectName("gridLayout")
        self.btnOpenFile = QtWidgets.QPushButton(self.frame)
        self.btnOpenFile.setObjectName("btnOpenFile")
        self.gridLayout.addWidget(self.btnOpenFile, 0, 0, 1, 1)
        self.btnNewFile = QtWidgets.QPushButton(self.frame)
        self.btnNewFile.setObjectName("btnNewFile")
        self.gridLayout.addWidget(self.btnNewFile, 1, 0, 1, 2)
        self.btnOpenRecent = QtWidgets.QPushButton(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.btnOpenRecent.sizePolicy().hasHeightForWidth())
        self.btnOpenRecent.setSizePolicy(sizePolicy)
        self.btnOpenRecent.setObjectName("btnOpenRecent")
        self.gridLayout.addWidget(self.btnOpenRecent, 0, 1, 1, 1)
        self.verticalLayout_3.addWidget(self.frame)
        spacerItem3 = QtWidgets.QSpacerItem(20, 5, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout_3.addItem(spacerItem3)
        self.frame_2 = QtWidgets.QFrame(StartupWidget)
        self.frame_2.setObjectName("frame_2")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.frame_2)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.btnSettings = QtWidgets.QPushButton(self.frame_2)
        self.btnSettings.setObjectName("btnSettings")
        self.gridLayout_2.addWidget(self.btnSettings, 0, 0, 1, 1)
        self.btnHelp = QtWidgets.QPushButton(self.frame_2)
        self.btnHelp.setObjectName("btnHelp")
        self.gridLayout_2.addWidget(self.btnHelp, 1, 0, 1, 1)
        self.verticalLayout_3.addWidget(self.frame_2)
        spacerItem4 = QtWidgets.QSpacerItem(20, 5, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout_3.addItem(spacerItem4)
        spacerItem5 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem5)
        self.btnQuit = QtWidgets.QPushButton(StartupWidget)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/pic/closehide.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnQuit.setIcon(icon)
        self.btnQuit.setObjectName("btnQuit")
        self.verticalLayout_3.addWidget(self.btnQuit)

        self.retranslateUi(StartupWidget)
        QtCore.QMetaObject.connectSlotsByName(StartupWidget)

    def retranslateUi(self, StartupWidget):
        _translate = QtCore.QCoreApplication.translate
        StartupWidget.setWindowTitle(_translate("StartupWidget", "Bibolamazi"))
        self.btnOpenFile.setText(_translate("StartupWidget", "Open Bibolamazi File"))
        self.btnNewFile.setText(_translate("StartupWidget", "Create New Bibolamazi File ..."))
        self.btnOpenRecent.setText(_translate("StartupWidget", "..."))
        self.btnSettings.setText(_translate("StartupWidget", "Settings"))
        self.btnHelp.setText(_translate("StartupWidget", "Help && Reference Browser"))
        self.btnQuit.setText(_translate("StartupWidget", "Quit"))


from . import bibolamazi_res_rc




# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2018 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr

import sys
import os
import os.path
import re
import logging
import textwrap
import codecs
from html import escape as htmlescape

import bibolamazi.init
from bibolamazi.core import bibolamazifile

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from .qtauto.ui_newbibolamazifiledialog import Ui_NewBibolamazifileDialog
from .sourcelisteditor import sanitize_bib_rel_path

logger = logging.getLogger(__name__)



# ------------------------------------------------------------------------------



class NewBibolamazifileDialog(QDialog):
    def __init__(self, parent=None):
        super(NewBibolamazifileDialog, self).__init__(parent)

        self.ui = Ui_NewBibolamazifileDialog()
        self.ui.setupUi(self)

        self.ui.txtSources.setHtml("""\
<!DOCTYPE HTML>
<html>
  <head>
    <style type="text/css">
p, li { white-space: pre-wrap; }
    </style>
  </head>
  <body>
    <p style="font-style:italic; color:#808080;">Please add source&nbsp;&nbsp;\N{RIGHTWARDS ARROW}</p>
  </body>
</html>
""")

        palette = self.ui.txtSources.palette()
        palette.setColor(QPalette.Base, palette.color(QPalette.Window))
        self.ui.txtSources.setPalette(palette)

        self.ui.btnSrcAdd.setVisible(False)
        self.ui.btnSrcClear.setVisible(False)
        self.ui.chkDuplicatesFilter.setVisible(False)

        self.ui.btnBack.setVisible(False)
        self.ui.btnSaveFinish.setVisible(False)

        self.ui.stk.setCurrentIndex(0)

        class _Ns: pass

        self.data = _Ns()
        self.data.src = []

        self.selected_save_fname = None


    bibolamaziFileCreated = pyqtSignal(str)

    def selectedSaveFileName(self):
        return self.selected_save_fname

    @pyqtSlot(int)
    def on_stk_currentChanged(self, pageno):
        self.ui.btnBack.setVisible(pageno > 0)
        if pageno == self.ui.stk.count()-1:
            self.ui.btnNext.setVisible(False)
            self.ui.btnSaveFinish.setVisible(True)
        else:
            self.ui.btnNext.setVisible(True)
            self.ui.btnSaveFinish.setVisible(False)

    @pyqtSlot()
    def on_btnNext_clicked(self):
        self.ui.stk.setCurrentIndex(self.ui.stk.currentIndex() + 1)

    @pyqtSlot()
    def on_btnBack_clicked(self):
        self.ui.stk.setCurrentIndex(self.ui.stk.currentIndex() - 1)

    @pyqtSlot()
    def on_btnSrcSet_clicked(self):
        
        fname, _filter = QFileDialog.getOpenFileName(self, 'Select BibTeX File', str(),
                                                     'BibTeX Files (*.bib);;All Files (*)')
        logger.debug("selected fname = %r", fname)
        if not fname:
            return

        self.data.src = [ fname ]
        self.updateSrcDisplay()

    @pyqtSlot()
    def on_btnSrcClear_clicked(self):
        
        self.data.src = [ ]
        self.updateSrcDisplay()

    @pyqtSlot()
    def on_btnSrcAdd_clicked(self):
        
        fnamelist, _filter = QFileDialog.getOpenFileNames(self, 'Select BibTeX File(s)', str(),
                                                          'BibTeX Files (*.bib);;All Files (*)')
        logger.debug("selected fname list = %r", fnamelist)
        if not fnamelist or not fnamelist[0]:
            return

        self.data.src += fnamelist
        self.updateSrcDisplay()

    def updateSrcDisplay(self):
        if self.data.src:
            html = "<html><head/><body>"
            if len(self.data.src) == 1:
                html += "<p>" + htmlescape(self.data.src[0]) + "</p>"
            else:
                html += "<ol>" + "".join([ "<li>"+htmlescape(x)+"</li>" for x in self.data.src ]) + "</ol>"
            html += "</body></html>"
            self.ui.txtSources.setHtml(html)
        else:
            self.ui.txtSources.setHtml(
                "<html><head/><body><p style=\"color: #808080; font-style:italic\">(no sources set)</p></body></html>"
            )


    @pyqtSlot(bool)
    def on_chkArxivUnpubIncludeTheses_toggled(self, on):
        self.ui.chkArxivPubIncludeTheses.setChecked(not on)
    
    @pyqtSlot(bool)
    def on_chkArxivPubIncludeTheses_toggled(self, on):
        self.ui.chkArxivUnpubIncludeTheses.setChecked(not on)


    @pyqtSlot()
    def on_btnSaveFinish_clicked(self):

        saveFileDialog = QFileDialog(self, "Create Bibolamazi File", str(),
                                     "Bibolamazi Files (*.bibolamazi.bib);;All Files (*)")
        if sys.platform.startswith('darwin'):
            # NOTE: BUG: OS X' file selection dialog won't understand the
            # .bibolamazi.bib extension. So use Qt's file dialog.
            saveFileDialog.setOptions(QFileDialog.DontUseNativeDialog)
        
        saveFileDialog.setDefaultSuffix("bibolamazi.bib")
        saveFileDialog.setAcceptMode(QFileDialog.AcceptSave)
        saveFileDialog.setFileMode(QFileDialog.AnyFile)
        if not saveFileDialog.exec_():
            return
        
        selectedfiles = saveFileDialog.selectedFiles()
        if not selectedfiles:
            return
        assert len(selectedfiles) == 1
        newfilename = selectedfiles[0]
            
        if (not newfilename):
            # cancelled maybe?
            return

        #if (os.path.exists(newfilename)):
        #    QMessageBox.critical(
        #        self, "File Exists",
        #        ("File %s already exists. Please select a file name to save to,"
        #         " without overwriting any existing file.")%(newfilename)
        #    )
        #    return self.on_btnSaveFinish_clicked()

        content = self.generateBibolamazifileConfig(ref_dir=os.path.dirname(newfilename))

        fullcontent = (bibolamazifile.TEMPLATE_HEADER + "\n" +
                       bibolamazifile.CONFIG_BEGIN_TAG + "\n"
                       "% " + content.replace("\n", "\n% ") + "\n" +
                       bibolamazifile.CONFIG_END_TAG + "\n\n")

        try:
            with codecs.open(newfilename, 'w', encoding='utf-8') as f:
                f.write(fullcontent)
        except Exception as e:
            QMessageBox.critical(self, "Error", "Error: Can't create file: %s"%(str(e)))
            return

        self.selected_save_fname = newfilename

        self.bibolamaziFileCreated.emit(newfilename)

        self.accept()

    
    def generateBibolamazifileConfig(self, ref_dir=None):

        srcmulti = self.ui.rdbtnMergeMultiple.isChecked()

        bibolamazi_config = """
%% This bibliography database uses BIBOLAMAZI:
%%
%%     https://github.com/phfaist/bibolamazi
%%
%% Use this file as your latex bibliography bibtex file, i.e.
%% \\bibliography{<this-file-name>.bibolamazi.bib}

%% This is the BIBOLAMAZI configuration section. Additional two leading
%% percent signs indicate comments within the configuration.

%% **** SOURCES ****

%% Specify where to get the bibtex entries from.  Alternatives may be specified
%% as several files for the same 'src:' command

"""

        if len(self.data.src):
            bibolamazi_config += "\n".join( [ 'src: '+sanitize_bib_rel_path(x, ref_dir=ref_dir)
                                              for x in self.data.src ] )
        else:
            bibolamazi_config += "src: <INSERT-SOURCE-PATH-HERE>"

        bibolamazi_config += """

%% Add additional sources here.

%% **** FILTERS ****

%% Specify filters here. Specify as many filters as you want, each with a
%% `filter:' directive. See also the "Help & Reference" information.

"""

        # merge duplicates?
        if srcmulti:
            if self.ui.chkDuplicatesFilter.isChecked():
                if self.ui.chkKeepOnlyUsed.isChecked():
                    bibolamazi_config += """\
%% Merge duplicates into a single bibtex entry and create aliases so that
%% \cite{...} commands may use either key interchangably
%%
%% Important: You must include "\\input{bibolamazi_dup_aliases.tex}" in the
%% preamble of your LaTeX document.
%%
%% Also, only keep entries that we actually use in the LaTeX document, and
%% discard any entries in the source bibtex files that we don't cite. NOTE: If
%% the LaTeX document has a different base name than the current bibolamazi
%% file, you need to set the -sJobname=documentname option here.
filter: duplicates -dMergeDuplicates
                   -dEnsureConflictKeysAreDuplicates
                   -sDupfile=bibolamazi_dup_aliases.tex
                   -dKeepOnlyUsed

"""
                else:
                    bibolamazi_config += """\
%% Merge duplicates into a single bibtex entry and create aliases so that
%% \cite{...} commands may use either key interchangably
%%
%% Important: You must include "\\input{bibolamazi_dup_aliases.tex}" in the
%% preamble of your LaTeX document.
filter: duplicates -dMergeDuplicates
                   -dEnsureConflictKeysAreDuplicates
                   -sDupfile=bibolamazi_dup_aliases.tex

"""
            else:
                if self.ui.chkKeepOnlyUsed.isChecked():
                    bibolamazi_config += """\
%% Don't merge duplicates, but make sure that any repeated entries between
%% different files *do* in fact refer to the same entry.
%%
%% Also, only keep entries that we actually use in the LaTeX document, and
%% discard any entries in the source bibtex files that we don't cite. NOTE: If
%% the LaTeX document has a different base name than the current bibolamazi
%% file, you need to set the -sJobname=documentname option here.
filter: duplicates -dMergeDuplicates=False
                   -dEnsureConflictKeysAreDuplicates
                   -dKeepOnlyUsed

"""
                else:
                    bibolamazi_config += """\
%% Don't merge duplicates, but make sure that any repeated entries between
%% different files *do* in fact refer to the same entry.
filter: duplicates -dMergeDuplicates=False -dEnsureConflictKeysAreDuplicates

"""

        else: # if not srcmulti
            if self.ui.chkKeepOnlyUsed.isChecked():
                bibolamazi_config += """\
%% Only keep entries that we actually use in the LaTeX document, and discard any
%% entries in the source bibtex files that we don't cite. NOTE: If the LaTeX
%% document has a different base name than the current bibolamazi file, you need
%% to set the -sJobname=documentname option here.
filter: only_used

"""

        if self.ui.chkFixesFilter.isChecked():
            fixesopts = []
            if self.ui.chkFixesGeneralSelection.isChecked():
                fixesopts += [ '-dFixSpaceAfterEscape', '-dRemoveTypeFromPhd', '-dRemoveFileField' ]
            if self.ui.chkFixesEncodeToLatex.isChecked():
                fixesopts += [ '-dEncodeUtf8ToLatex' ]

            bibolamazi_config += ("""\
%%%% Apply some general fixes & hacks
filter: fixes %(fixesoptstr)s

"""%dict(fixesoptstr="\n              ".join(fixesopts)))


        if self.ui.chkArxivUnpublished.isChecked():
            bibolamazi_config += """\
%% Format arxiv references for *unpublished* entries."""
            if self.ui.chkArxivUnpubIncludeTheses.isChecked():
                thesespublishedopt = '-dThesesCountAsPublished=False'
            else:
                thesespublishedopt = '-dThesesCountAsPublished'
            if self.ui.rdbtnArxivUnpub_eprint.isChecked():
                bibolamazi_config += ("""
%%%% Include a 'eprint = {...}' field with arXiv identifier:
filter: arxiv -sUnpublishedMode=eprint
              -sMode=none
              %(thesespublishedopt)s
              -dWarnJournalRef

"""%dict(thesespublishedopt=thesespublishedopt))
            elif self.ui.rdbtnArxivUnpub_unpubnote.isChecked():
                bibolamazi_config += ("""
%%%% Mark bibtex entry as type @unpublished{...}, and include a bibtex field
%%%% 'note = {arXiv:...}' with the arXiv identifier:
filter: arxiv -sUnpublishedMode=unpublished-note
              -sMode=none
              %(thesespublishedopt)s
              -dWarnJournalRef

"""%dict(thesespublishedopt=thesespublishedopt))
            else:
                logger.warning("arxiv unpublished was checked, but no explicit mode was set")
        
        if self.ui.chkArxivPublished.isChecked():
            bibolamazi_config += """\
%% Format arxiv references for entries considered as *published*."""
            if self.ui.chkArxivPubIncludeTheses.isChecked():
                thesespublishedopt = '-dThesesCountAsPublished'
            else:
                thesespublishedopt = '-dThesesCountAsPublished=False'
            if self.ui.rdbtnArxivPub_eprint.isChecked():
                bibolamazi_config += ("""
%%%% Include a 'eprint = {...}' field with arXiv identifier:
filter: arxiv -sMode=eprint
              -sUnpublishedMode=none
              %(thesespublishedopt)s

"""%dict(thesespublishedopt=thesespublishedopt))
            elif self.ui.rdbtnArxivPub_unpubnote.isChecked():
                bibolamazi_config += ("""
%%%% Include a 'note = {arXiv:...}' field with arXiv identifier:
filter: arxiv -sMode=note
              -sUnpublishedMode=none
              %(thesespublishedopt)s

"""%dict(thesespublishedopt=thesespublishedopt))
            elif self.ui.rdbtnArxivPub_strip.isChecked():
                bibolamazi_config += ("""
%%%% Strip out arxiv information entirely
filter: arxiv -sMode=strip
              -sUnpublishedMode=none
              %(thesespublishedopt)s

"""%dict(thesespublishedopt=thesespublishedopt))
            else:
                logger.warning("arxiv published was checked, but no explicit mode was set")
        

        if self.ui.chkUrlFilter.isChecked():
            urlopts = []
            if self.ui.chkUrlStrip.isChecked():
                urlopts.append('-dStrip')
            else:
                if self.ui.chkUrlStripAllIfDoiOrArxiv.isChecked():
                    urlopts.append('-dStripAllIfDoiOrArxiv')
                if self.ui.chkUrlStripDoiArxiv.isChecked():
                    urlopts += [ '-dStripDoiUrl', '-dStripArxivUrl' ]
                if self.ui.chkUrlKeepFirstUrlOnly.isChecked():
                    urlopts.append('-dKeepFirstUrlOnly')
            bibolamazi_config += ("""\
%%%% Clean up URLs in bibtex entries (the 'url = {...}' field)
filter: url %(urloptstr)s

"""%dict(urloptstr="\n            ".join(urlopts)))

        # always add the order_entries filter
        bibolamazi_config += """\
%% Order the entries in the final bibtex output. This has usually no effect on
%% the final bibliography formatting, but it avoids large file differences at
%% each run of bibolamazi (useful for instance when using git)
filter: orderentries -sOrder=alphabetical

"""

        return bibolamazi_config




# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr
import sys
def to_native_str(x): return x.encode('utf-8') if sys.version_info[0] <= 2 else x
def from_native_str(x): return x.decode('utf-8') if sys.version_info[0] <= 2 else x
from imp import reload


from html import escape as htmlescape
import sys
import os
import os.path
import re
import textwrap
import shlex
import logging
import importlib

import bibolamazi.init
from bibolamazi.core import main as bibolamazimain
from bibolamazi.core import blogger
from bibolamazi.core.blogger import logger
from bibolamazi.core import bibolamazifile
from bibolamazi.core import butils
from bibolamazi.core.butils import BibolamaziError
from bibolamazi.core.bibfilter import factory as filterfactory

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from . import uiutils
from .uiutils import ContextAttributeSetter
from . import buttontabsmanager
from .bibconfigsynthigh import BibolamaziConfigSyntaxHighlighter
from .favorites import FavoriteCmd, FavoritesModel, FavoritesItemDelegate, FavoritesOverBtns
from . import filterinstanceeditor
from . import filterpackagepatheditor
from . import settingswidget
from . import helpbrowser
from . import searchwidget

from .qtauto.ui_openbibfile import Ui_OpenBibFile

logger = logging.getLogger(__name__)



def bibolamazi_error_html(errortxt, wrap_pre=True):

    def a_link(m):
        return "<a href=\"action:/goto-bibolamazi-file-line/%d\">%s</a>" %(
            int(m.group('lineno')),
            m.group()
            )

    errortxt = str(htmlescape(errortxt, quote=True))
    errortxt = re.sub(r'@:.*line\s+(?P<lineno>\d+)', a_link, errortxt)
    try:
        # if wrap_pre = (start_tag, end_tag)
        return wrap_pre[0] + errortxt + wrap_pre[1]
    except (TypeError,IndexError):
        pass
    if wrap_pre:
        # if wrap_pre = True
        return ("<pre style=\"white-space: pre-wrap\">"+errortxt+"</pre>")
    return errortxt



class PreformattedHtml(object):
    def __init__(self, html):
        self.html = str(html)

    def getHtml(self):
        return self.html

    def __str__(self):
        return to_native_str(self.html)
    def __unicode__(self):
        return self.html


class LogToHtmlQtSignal(QObject, logging.Handler):
    #
    # Handler to generate Qt signals from log messages.
    #
    # IMPORTANT: Additionally, this class interacts with RunBibolamaziThread to
    # do some internal event processing by hacking into the logging mechanism.
    #

    def __init__(self, parent, threadparent=None):
        super(LogToHtmlQtSignal, self).__init__(parent)
        logging.Handler.__init__(self)

        self.threadparent = threadparent

    clearLog = pyqtSignal()
    logHtml = pyqtSignal(str)

    def doclear(self):
        self.clearLog.emit()

    def dolog(self, txt, levelno=logging.INFO):

        try:
            html = txt.getHtml() # in case of a PreformattedHtml instance
        except AttributeError:
            html = str(htmlescape(txt)) # in case of a simple plain text string

        sty = ''
        if levelno == logging.ERROR or levelno == logging.CRITICAL:
            sty = "color: #ff0000; font-weight: bold;"
        elif levelno == logging.WARNING:
            sty = "color: rgb(150,80,0); font-weight: bold;"
        elif levelno == logging.INFO:
            sty = "font-weight: normal;" # default color #"color: #000000; font-weight: normal;"
        elif levelno == logging.DEBUG or levelno == blogger.LONGDEBUG:
            sty = "color: #7f7f7f; font-weight: normal;"
        else:
            # unknown level
            sty = "color: #7f7f7f; font-weight: normal;"

        sty += "white-space: pre;"

        self.logHtml.emit("<span style=\"%s\">%s\n</span>"%(sty, html))

    def emit(self, record):
        self.dolog(self.format(record), record.levelno)

        if hasattr(self.threadparent, 'my_process_events'):
            self.threadparent.my_process_events()




class LogToGuiContextManager(object):
    def __init__(self, logqtsig, bibolamazi_fname, thelogger=logging.getLogger()):
        self.ch = logqtsig
        self.ch.setLevel(logging.NOTSET) # propagate all messages

        self.bibolamazi_fname = bibolamazi_fname

        # create formatter and add it to the handlers
        self.formatter = blogger.ConditionalFormatter('%(message)s',
                                                      DEBUG='-- %(message)s',
                                                      LONGDEBUG='  -- %(message)s',
                                                      WARNING='WARNING: %(message)s',
                                                      ERROR=self._fmt_error,
                                                      CRITICAL='CRITICAL: %(message)s')
        self.ch.setFormatter(self.formatter)

        self.logger = thelogger

    def _fmt_error(self, d):
        html = bibolamazi_error_html(
            '%(message)s' % d, # d is the dictionary with the record information
            wrap_pre=('<span style="white-space: pre">', '</span>')
        )
        html = "<br />".join(['ERROR: %s'%(line) for line in html.split('\n')])
        return PreformattedHtml(html=html)

    def __enter__(self):
        # add the handlers to the logger
        self.logger.addHandler(self.ch)
        self.ch.doclear()
        return self
        

    def __exit__(self, type, value, traceback):
        # clean-up
        self.logger.removeHandler(self.ch)





class RunBibolamaziThread(QThread):
    def __init__(self, parent):
        super(RunBibolamaziThread, self).__init__(parent)
        self.setObjectName("run-bibolamazi-thread")

        self._orig_except_hook = sys.excepthook
        sys.excepthook = self._py_except_hook

        self.keyboardinterrupt_requested = False

        self.busy = False # accessed directly by RunBibolamazi objects

    def getOwnRunBibolamazi(self):
        r = RunBibolamazi(self)
        r.moveToThread(self)
        return r

    @pyqtSlot()
    def doQuit(self):
        self.quit()
        self.wait()
        sys.excepthook = self._orig_except_hook

    def _py_except_hook(self, etype, evalue, tb):
        if etype is KeyboardInterrupt:
            self.keyboardinterrupt_requested = True
            return
        self._orig_except_hook(etype, evalue, tb)

    def my_process_events(self):
        if self.keyboardinterrupt_requested:
            self.keyboardinterrupt_requested = False
            raise KeyboardInterrupt

        QCoreApplication.processEvents()


global_run_bibolamazi_thread = None
def global_run_bibolamazi_thread_instance():
    global global_run_bibolamazi_thread
    if global_run_bibolamazi_thread is None:
        global_run_bibolamazi_thread = RunBibolamaziThread(QApplication.instance())
        global_run_bibolamazi_thread.start()
    return global_run_bibolamazi_thread


class RunBibolamazi(QObject):
    def __init__(self, threadparent):
        super(RunBibolamazi, self).__init__(None) # important: no parent (for moveToThread())

        self.threadparent = threadparent

        self.logqtsig = LogToHtmlQtSignal(self, threadparent)
        self.logqtsig.clearLog.connect(self.clearLog)
        self.logqtsig.logHtml.connect(self.logHtml)

        self._busy = False


    busy = pyqtSignal(bool)
    bibolamaziDone = pyqtSignal()
    bibolamaziDoneError = pyqtSignal(str)

    clearLog = pyqtSignal()
    logHtml = pyqtSignal(str)

    def isBusy(self):
        return self._busy
    
    @pyqtSlot(str, int, list)
    def runBibolamazi(self, bibolamazifilename, verbosity_level, reload_filter_path):

        # reload_filter_path is a list of 2-tuples (fpname, fpdir)

        if self.threadparent.busy:
            logger.warning("Can't run bibolamazi twice simultaneously")
            return

        logger.debug("RunBibolamazi.runBibolamazi(): bibolamazifilename=%r, verbosity_level=%r, "
                     "reload_filter_path=%r. thread id=%r",
                     bibolamazifilename, verbosity_level, reload_filter_path,
                     QThread.currentThread().objectName())

        self._busy = True
        self.threadparent.busy = True
        self.busy.emit(True)

        try:

            rootlogger = logging.getLogger()
            loggerleveltoset = bibolamazimain.verbosity_logger_level(verbosity_level)
            with ContextAttributeSetter(
                    (lambda : rootlogger.level, rootlogger.setLevel, loggerleveltoset), ):
                with LogToGuiContextManager(logqtsig=self.logqtsig,
                                            bibolamazi_fname=bibolamazifilename) as log2sig:
                    try:
                        # TODO: we should reload all relevant packages here, not
                        # in the other method below

                        self._reload_filterpkg_modules(reload_filter_path)

                        bibolamazimain.run_bibolamazi(bibolamazifile=bibolamazifilename)

                        self.logqtsig.dolog(" --> Finished successfully. <--")
                        self.bibolamaziDone.emit()
                    except butils.BibolamaziError as e:
                        logger.error(str(e))
                        self.logqtsig.dolog(" --> Finished with errors. <--")
                        self.bibolamaziDoneError.emit(str(e))
                    except KeyboardInterrupt as e:
                        logger.error("*** interrupted ***")
                        self.logqtsig.dolog(" --> Stop: Interrupted. <--")
                        self.bibolamaziDoneError.emit("Bibolamazi interrupted")
                    except Exception as e:
                        stre = str(e)
                        logger.error(stre)
                        import traceback
                        logger.error(traceback.format_exc())
                        self.logqtsig.dolog(" --> EXCEPTION <--")
                        self.bibolamaziDoneError.emit(stre)

        finally:
            self.busy.emit(False)
            self._busy = False
            self.threadparent.busy = False


    def _reload_filterpkg_modules(self, reload_filter_path):
        #
        # reload all relevant local filter packages, in case the filter packages
        # have changed.
        #
        # reload_filter_path is a list of 2-tuples (fpname, fpdir)
        #
        allmodules = sorted(sys.modules.keys(), reverse=True) # so that "pkg.submodule" appears before "pkg"
        for pkgname, pkgpath in reload_filter_path:
            logger.debug("Inspecting user filter package `%s` to reload modules ...", pkgname)
            for modname in allmodules:
                if modname.startswith(pkgname):
                    mod = sys.modules[modname]
                    logger.debug("Reloading module `%s` (%r) in user filter package `%s`",
                                 modname, mod, pkgname)
                    origpath = sys.path
                    try:
                        sys.path = [pkgpath] + origpath
                        reload(mod)
                    finally:
                        sys.path = origpath

    #@pyqtSlot()
    #def _process_some_events(self):
    #    # this processes events for current thread (from Qt docs)
    #    QCoreApplication.processEvents()



                        

class OpenBibFile(QWidget):

    requestHelpTopic = pyqtSignal('QString')

    fileClosed = pyqtSignal()
    

    def __init__(self):
        super(OpenBibFile, self).__init__()

        self.ui = Ui_OpenBibFile()
        self.ui.setupUi(self)

        self.bibolamazi_thread = global_run_bibolamazi_thread_instance()

        self.run_bibolamazi = self.bibolamazi_thread.getOwnRunBibolamazi()
        self.run_bibolamazi.busy.connect(self._run_busy)
        self.run_bibolamazi.bibolamaziDone.connect(self._run_bibolamaziDone)
        self.run_bibolamazi.bibolamaziDoneError.connect(self._run_bibolamaziDoneError)
        self.run_bibolamazi.clearLog.connect(self._run_clearLog)
        self.run_bibolamazi.logHtml.connect(self._run_logHtml)

        self.requestRunBibolamazi.connect(self.run_bibolamazi.runBibolamazi)

        self.resize(QSize(1200,800))
        self.ui.splitEditConfig.setSizes([700,500])

        # connect buttons as "tabs"
        self.btabs = buttontabsmanager.ButtonTabsManager(self.ui.tabs, self)
        self.btabs.registerButton(self.ui.btnPageEditConfig, self.ui.tabPageEditFilters)
        self.btabs.registerButton(self.ui.btnPageFileInfo, self.ui.tabPageFileInfo)
        self.btabs.registerButton(self.ui.btnPageRunMessages, self.ui.tabPageRunMessages)

        # also in info page
        self.btabsinfo = buttontabsmanager.ButtonTabsManager(self.ui.stkInfo, self)
        self.btabsinfo.registerButton(self.ui.btnInfoPageHeader, self.ui.pageInfoHeader)
        self.btabsinfo.registerButton(self.ui.btnInfoPageEntries, self.ui.pageInfoEntries)

        self.ui.txtConfig.setWordWrapMode(QTextOption.WrapAnywhere)
        self.ui.txtLog.setWordWrapMode(QTextOption.WrapAnywhere)
        self.ui.txtBibEntries.setWordWrapMode(QTextOption.WrapAnywhere)

        p = QPalette(self.ui.stackEditTools.palette())
        c = p.color(QPalette.Active, QPalette.Window)
        p.setColor(QPalette.Window, QColor(int(c.red()*0.9), int(c.green()*0.9), int(c.blue()*0.9)))
        self.ui.stackEditTools.setPalette(p)
        self.ui.stackEditTools.setAutoFillBackground(True)

        font = settingswidget.get_typewriter_font(self.ui.txtConfig)
        self.ui.txtConfig.setFont(font)
        self.ui.txtLog.setFont(font)
        self.ui.txtBibEntries.setFont(font)

        self.searchConfigManager = searchwidget.SearchTextEditManager(
            self.ui.searchConfigWidget,
            self.ui.txtConfig
        )

        self.searchPrevBibEntriesManager = searchwidget.SearchTextEditManager(
            self.ui.searchPrevBibEntries,
            self.ui.txtBibEntries
        )

        self.syntHighlighter = BibolamaziConfigSyntaxHighlighter(self.ui.txtConfig)

        self._favorites_overbtn = FavoritesOverBtns(self.ui.treeFavorites)
        self._favorites_overbtn.insertCommand.connect(self._insert_new_cmd)

        self.ui.filterInstanceEditor.requestAddToFavorites.connect(self.add_favorite_cmd)
        self.ui.sourceListEditor.requestAddToFavorites.connect(self.add_favorite_cmd)
        self.ui.sourceListEditor.requestAddSourceList.connect(self.on_btnAddSourceList_clicked)

        self.ui.btnSelectedMultipleAddFavorite.clicked.connect(self.add_favorite_selection)

        self.bibolamaziFileName = None
        self.bibolamaziFile = None

        # to implement delayed reparsing of the current textedit contents
        self.reparseTimer = QTimer(self)
        self.reparseTimer.timeout.connect(self._reparse_textedit)
        
        # to implement delayed full reload of the file
        self.updateTimer = QTimer(self)
        self.updateTimer.timeout.connect(self.reloadFile)
        

        self._flag_modified_externally = False

        self.fwatcher = QFileSystemWatcher(self)
        self.fwatcher.fileChanged.connect(self.fileModifiedExternally)
        #self.delayedUpdateFileContents)

        self.shortcuts = [
            QShortcut(QKeySequence('Ctrl+W'), self, self.close, self.close),
            QShortcut(QKeySequence('Ctrl+S'), self, self.saveToFile, self.saveToFile),
            QShortcut(QKeySequence(Qt.CTRL|Qt.Key_Return), self, self.runBibolamazi, self.runBibolamazi),
            ]

        self._ignore_change_for_edittools = False
        self._needs_update_txtbibentries = False
        self._set_modified(False)

        logger.longdebug("finished OpenBibFile constructor!")

    def setFavoriteCmdsList(self, favoriteCmdsList):
        self.favoriteCmdsList = favoriteCmdsList

        self._favorites_model = FavoritesModel(favcmds=self.favoriteCmdsList, parent=self)
        self.ui.treeFavorites.setModel(self._favorites_model)

        self._favorites_delegate = FavoritesItemDelegate(parent=self)
        self.ui.treeFavorites.setItemDelegate(self._favorites_delegate)
        

    def setOpenFile(self, filename):
        if self.bibolamaziFileName:
            self.fwatcher.removePath(self.bibolamaziFileName)
            
        self.bibolamaziFileName = filename
        self.bibolamaziFile = None

        ok = self.reloadFile()

        if self.bibolamaziFileName:
            self.fwatcher.addPath(self.bibolamaziFileName)

        return ok


    def hasUnsavedModifications(self):
        return self._modified

    def fileName(self):
        return self.bibolamaziFileName

    # use close() to close the widget and file
    def closeEvent(self, closeEvent):

        # refuse to close if the bibolamazi thread is busy for us
        if self.run_bibolamazi.isBusy():
            logger.debug("Bibolamazi is running, can't close window")
            closeEvent.ignore()
            return

        logger.debug("closeEvent(): cleaning up.")
        
        if self._modified:
            ans = QMessageBox.question(self, 'Save Changes',
                                       "Save changes to file `%s'?" %(self.bibolamaziFileName,),
                                       QMessageBox.Save|QMessageBox.Discard|QMessageBox.Cancel,
                                       QMessageBox.Save)
            if ans == QMessageBox.Save:
                # save first (no need for updating), then proceed with close
                self.saveToFile(noupdate=True)
                
            if ans == QMessageBox.Cancel:
                # ignore the event
                closeEvent.ignore()
                return
        
        if self.bibolamaziFileName:
            self.fwatcher.removePath(self.bibolamaziFileName)
            self.bibolamaziFileName = None
        if self.bibolamaziFile:
            self.bibolamaziFile = None

        QMetaObject.invokeMethod(self, 'fileClosed', Qt.QueuedConnection)
        
        return super(OpenBibFile, self).closeEvent(closeEvent)


    @pyqtSlot()
    def saveToFile(self, noupdate=False):
        if not self.bibolamaziFile:
            QMessageBox.critical(self, "No File!", "No file to save to!")
            return

        config_data = str(self.ui.txtConfig.toPlainText())

        # ignore changes caused by ourselves
        self.fwatcher.blockSignals(True)
        
        # change the config block.
        self.bibolamaziFile.setConfigData(config_data)
        self.bibolamaziFile.saveRawToFile()

        self._set_modified(False)

        # reload file.
        if not noupdate:
            self.delayedUpdateFileContents()

    @pyqtSlot()
    def fileModifiedExternally(self):
        logger.debug("File modified externally!!")
        self._flag_modified_externally = True
        #self.delayedUpdateFileContents

    @pyqtSlot()
    def delayedUpdateFileContents(self):
        if self.updateTimer.isActive():
            self.updateTimer.stop()

        self.ui.txtConfig.setReadOnly(True)

        self.updateTimer.setInterval(1000)
        self.updateTimer.setSingleShot(True)
        self.updateTimer.start()



    def _set_win_state(self, file_is_loaded, errormessagehtml=None, ifOkSetConfigTab=False):

        # enable/disable these GUI widgets/buttons
        
        self.ui.tabPageEditFilters.setEnabled(file_is_loaded)
        self.ui.pageBibEntries.setEnabled(file_is_loaded)
        self.ui.tabPageRunMessages.setEnabled(file_is_loaded)

        self.ui.btnSave.setEnabled(file_is_loaded)
        self.ui.btnGo.setEnabled(file_is_loaded)
        self.ui.btnInterrupt.setVisible(False)
        

        if not file_is_loaded:
            # disabled state

            self.ui.tabs.setCurrentWidget(self.ui.tabPageRunMessages)
            self.ui.txtParseErrorMessages.setVisible(True)
            self.ui.txtParseErrorMessages.setText(errormessagehtml)

            self.ui.txtInfo.setText("<h3 style=\"color: rgb(127,0,0)\">parse error in file.</h3>")

            with ContextAttributeSetter( (self.ui.txtConfig.signalsBlocked,
                                          self.ui.txtConfig.blockSignals, True) ):
                self.ui.txtConfig.setPlainText("")

            self.ui.lblFileName.setText("<no file loaded>")

            self.setWindowFilePath(str())
            self.setWindowTitle("<no file loaded>")
            self.setWindowIcon(QIcon(':/pic/file.png'))

            self._modified = False

            return

        # normal, enabled state
        self.ui.lblFileName.setText(self.bibolamaziFileName)

        self.ui.txtParseErrorMessages.setVisible(False)

        self.setWindowFilePath(self.bibolamaziFileName)
        self.setWindowTitle(os.path.basename(self.bibolamaziFileName))
        self.setWindowIcon(QIcon(':/pic/file.png'))

        if ifOkSetConfigTab:
            self.ui.tabs.setCurrentWidget(self.ui.tabPageEditFilters)
        self.ui.txtConfig.setFocus()


    @pyqtSlot()
    def reloadFile(self):
        logger.debug("reloading file")

        self._flag_modified_externally = False

        # we may be sensitive again to external changes again
        if self.fwatcher.signalsBlocked():
            self.fwatcher.blockSignals(False)

        if not self.bibolamaziFileName:
            with ContextAttributeSetter( (self.ui.btnGo.isEnabled, self.ui.btnGo.setEnabled, False) ):
                self.ui.txtInfo.setText("<h3 style=\"color: rgb(127,0,0)\">no file loaded.</h3>")
                self.ui.txtConfig.setPlainText("")
            return False

        if not self.bibolamaziFile:
            self.bibolamaziFile = bibolamazifile.BibolamaziFile()

        try:

            self.bibolamaziFile.load(self.bibolamaziFileName, to_state=bibolamazifile.BIBOLAMAZIFILE_READ)

        except butils.BibolamaziError as e:
            self.bibolamaziFile = None
            QMessageBox.critical(self, "Load Error", u"Error loading file: %s" %(str(e)))
            self._set_win_state(False,
                                "<h3 style=\"color: rgb(127,0,0)\">Error reading file.</h3>\n"
                                + bibolamazi_error_html(str(e)))
            return False

        self._set_win_state(True)

        # all set, clear any read-only flags on the txtConfig
        self.ui.txtConfig.setReadOnly(False)

        has_cache = os.path.exists(os.path.join(self.bibolamaziFile.fdir(),
                                                self.bibolamaziFileName + '.bibolamazicache'))
        # if there is a cache file, then offer to delete it
        self._set_ui_delete_cache_btn(has_cache)

        with ContextAttributeSetter( (self.ui.txtConfig.signalsBlocked,
                                      self.ui.txtConfig.blockSignals, True) ):
            cursorpos = self.ui.txtConfig.textCursor().position()
            self.ui.txtConfig.setPlainText(self.bibolamaziFile.configData())
            cur = self.ui.txtConfig.textCursor()
            cur.setPosition(cursorpos)
            self.ui.txtConfig.setTextCursor(cur)

        # now, try to further parse the config
        self._bibolamazifile_reparse()

        self._set_modified(False)

        self._needs_update_txtbibentries = True

        if self.bibolamaziFile.getLoadState() == bibolamazifile.BIBOLAMAZIFILE_PARSED:
            self._display_header()

        logger.debug("file contents updated!")
        return True


    @pyqtSlot()
    def on_btnDeleteCache_clicked(self):
        cachefile = os.path.join(self.bibolamaziFile.fdir(),
                                 self.bibolamaziFileName + '.bibolamazicache')
        if not os.path.exists(cachefile):
            QMessageBox.warning(self, "No cache file",
                                "There is no cache file to remove (%s)"%(cachefile))
            return

        logger.debug("removing cache file %r", cachefile)
        os.remove(cachefile)

        self._set_ui_delete_cache_btn(False)
        
    def _set_ui_delete_cache_btn(self, is_on):
        if is_on:
            self.ui.btnDeleteCache.setText("Delete cache")
            self.ui.btnDeleteCache.setEnabled(True)
        else:
            self.ui.btnDeleteCache.setText("(no cache file)")
            self.ui.btnDeleteCache.setEnabled(False)
            

    def _display_header(self):

        rawheader = self.bibolamaziFile.rawHeader().strip()
        if rawheader:
            fileinfohtml = ("<h2>Raw Header</h2>\n"
                            "<p class=\"shadow\">The top section of the bibolamazi file "
                            "is ignored by bibolamazi.  Whatever bibtex entries listed here "
                            "will not be affected by bibolamazi filters and will be retained "
                            "as is at the top of the file.  These bibtex entries are seen by "
                            "latex as regular entries that have not been filtered by bibolamazi.  "
                            "This portion of the file cannot be edited here; use your favorite "
                            "text editor to edit.</p>"
                            "<pre class=\"small\">" + htmlescape(rawheader) + "</pre>")

        dark_mode = uiutils.is_dark_mode(self)

        thehtml = textwrap.dedent('''\
        <!DOCTYPE HTML>
        <html>
          <head>
            <style type="text/css">
              %(basecss)s
              .source { margin: 0.5em 0px 0px 0px; }
              .filter { margin: 0.3em 0px 0px 0px; }
              .filterdescription { font-style: italic; margin-left: 2.5em; }
            </style>
          </head>
          <body>
            %(content)s
          </body>
        </html>''') % {
            'basecss': helpbrowser.getCssHelpStyle(dark_mode=dark_mode),
            'content': helpbrowser.wrapInHtmlContentContainer(fileinfohtml, width=800)
        }
        self.ui.txtInfo.setHtml(thehtml)
        

    @pyqtSlot(bool)
    def on_btnInfoPageEntries_toggled(self, on):
        logger.debug("on_btnInfoPageEntries_toggled(%r)", on)
        if on: # self.ui.stkInfo.widget(index) is self.ui.pageInfoEntries:
            logger.debug("redisplaying bib entries")
            if self._needs_update_txtbibentries:
                self.ui.txtBibEntries.setPlainText(self.bibolamaziFile.rawRest())

    @pyqtSlot(bool)
    def on_btnPageFileInfo_toggled(self, on):
        logger.debug("on_btnPageFileInfo_toggled(%r)", on)
        if on and self.ui.btnInfoPageEntries.isChecked():
            logger.debug("redisplaying bib entries")
            if self._needs_update_txtbibentries:
                self.ui.txtBibEntries.setPlainText(self.bibolamaziFile.rawRest())


    @pyqtSlot()
    def on_btnGo_clicked(self):
        self.ui.txtConfig.setReadOnly(True) # enter read-only state because
                                            # otherwise changes would get
                                            # overwritten right away
        self.runBibolamazi()

    @pyqtSlot()
    def on_btnInterrupt_clicked(self):
        self.ui.btnInterrupt.setEnabled(False) # not twice in a row
        import signal

        sig = signal.SIGINT
        if sys.platform.startswith('win'):
            logger.warning("Interrupt does not work on Windows (not implemented).")
            return
            # doesn't work????!?
            #logger.debug("Sending Ctrl+C ...")
            #sig = signal.CTRL_C_EVENT
            
        os.kill(os.getpid(), sig)


    requestRunBibolamazi = pyqtSignal(str, int, list)
    
    @pyqtSlot(bool)
    def _run_busy(self, busy):

        # block notifications for file contents updates that we generate ourselves...
        self.fwatcher.blockSignals(busy)

        if sys.platform.startswith('win'):
            self.ui.btnGo.setEnabled(not busy)
            self.ui.btnInterrupt.setVisible(False)
        else:
            self.ui.btnInterrupt.setVisible(busy)
            self.ui.btnInterrupt.setEnabled(busy)
            self.ui.btnGo.setVisible(not busy)

        if busy:
            self._busy_before_curtab = self.ui.tabs.currentWidget()
            self.ui.tabs.setCurrentWidget(self.ui.tabPageRunMessages)


    @pyqtSlot()
    def _run_bibolamaziDone(self):
        # if we're done with success, go back to previous tab
        if self._busy_before_curtab is not None:
            self.ui.tabs.setCurrentWidget(self._busy_before_curtab)
            self._busy_before_curtab = None
        self.delayedUpdateFileContents()

    @pyqtSlot(str)
    def _run_bibolamaziDoneError(self, errstr):
        QMessageBox.warning(self, "Bibolamazi error", errstr)
        self._busy_before_curtab = None
        self.delayedUpdateFileContents()

    @pyqtSlot()
    def _run_clearLog(self):
        self.ui.txtLog.clear()

    @pyqtSlot(str)
    def _run_logHtml(self, html):
        cur = QTextCursor(self.ui.txtLog.document())
        cur.movePosition(QTextCursor.End)
        cur.insertHtml(html)


    @pyqtSlot()
    def runBibolamazi(self):
        if self._modified:
            yn = QMessageBox.question(self, "Save Changes?", "Save changes to your config?",
                                      QMessageBox.Save|QMessageBox.Cancel, QMessageBox.Save)
            if yn == QMessageBox.Cancel:
                return

            # save our config
            self.saveToFile()
            
        if not self.bibolamaziFileName:
            QMessageBox.critical(self, "No open file", "No file selected!")
            return

        if not self.bibolamaziFile:
            QMessageBox.critical(self, "No open file", "No file is open!")
            return

        verbosity_level = self.ui.cbxVerbosity.currentIndex()

        self.ui.txtParseErrorMessages.setVisible(False)

        reload_filter_path = list(self.bibolamaziFile.filterPath().items())

        self.requestRunBibolamazi.emit(self.bibolamaziFileName, verbosity_level,
                                       reload_filter_path)


    @pyqtSlot(int)
    def gotoConfigLineNo(self, configlineno):
        self.ui.tabs.setCurrentWidget(self.ui.tabPageEditFilters)
        cur = QTextCursor(self.ui.txtConfig.document().findBlockByNumber(configlineno-1))
        self.ui.txtConfig.setTextCursor(cur)


    @pyqtSlot()
    def on_btnSave_clicked(self):
        self.saveToFile()

    @pyqtSlot()
    def on_btnRefresh_clicked(self):
        if self._modified:
            yn = QMessageBox.question(self, "Really revert?",
                                      self.tr("You are about to revert the file, but you "
                                              "have unsaved changes. Do you really want "
                                              "to abandon your changes and revert the "
                                              "file to disk?"),
                                      QMessageBox.Ok|QMessageBox.Cancel, QMessageBox.Cancel)
            if yn != QMessageBox.Ok:
                return

        self.reloadFile()


    @pyqtSlot(QUrl)
    def on_txtInfo_anchorClicked(self, url):
        self._open_anchor(url)

    @pyqtSlot(QUrl)
    def on_txtParseErrorMessages_anchorClicked(self, url):
        self._open_anchor(url)

    @pyqtSlot(QUrl)
    def on_txtLog_anchorClicked(self, url):
        self._open_anchor(url)

    def _open_anchor(self, url):
        logger.debug("Link clicked: %r", url.toString())

        if url.scheme() == "helptopic":
            self.requestHelpTopic.emit(url.path())
            return

        if url.scheme() == "action":
            m = re.match(r'/goto-config-line/(?P<lineno>\d+)', url.path())
            if m:
                self.gotoConfigLineNo(int(m.group('lineno')))
                return

            m = re.match(r'/goto-bibolamazi-file-line/(?P<lineno>\d+)', url.path())
            if m:
                if (self.bibolamaziFile is None or
                    self.bibolamaziFile.getLoadState() < bibolamazifile.BIBOLAMAZIFILE_READ):
                    logger.warning("No bibolamazifile open!")
                    return
                configlineno = self.bibolamaziFile.configLineNo(int(m.group('lineno')))
                self.gotoConfigLineNo(configlineno)
                return

            logger.warning("Unknown action: %s", str(url.toString()))
            return

        logger.debug("Opening URL %r", str(url.toString()))
        QDesktopServices.openUrl(url)
        

    @pyqtSlot(bool)
    def _set_modified(self, modif):
        self._modified = modif
        self.ui.btnSave.setEnabled(self._modified)

    @pyqtSlot()
    def on_txtConfig_textChanged(self):
        logger.debug("text changed! fwatcher signals blocked=%d", self.fwatcher.signalsBlocked())

        if self._flag_modified_externally:
            logger.debug("Modified externally!")
            
            revertbtn = None
            box = None
            if not self._modified:
                # file was modified externally, but we have no local changes.
                box = QMessageBox(QMessageBox.Warning, "File Modified Externally",
                                  "The File was modified externally. Revert and reload it?")
                revertbtn = box.addButton("Revert", QMessageBox.AcceptRole)
                box.addButton(QMessageBox.Cancel)
                box.setDefaultButton(revertbtn)
            else:
                # file was modified externally, but we have LOCAL UNSAVED CHANGES
                box = QMessageBox(QMessageBox.Warning, "File Modified Externally",
                                  "The File was modified externally, but you have unsaved changes. "
                                  "Keep your changes and ignore the external modification, or "
                                  "revert it to version on disk?")
                revertbtn = box.addButton("Revert", QMessageBox.AcceptRole)
                box.addButton("Keep changes", QMessageBox.RejectRole)
                box.setDefaultButton(revertbtn)

            box.exec_()
            clickedbtn = box.clickedButton()
            if clickedbtn == revertbtn:
                logger.debug("Reverting!!")
                # reload the file
                self.reloadFile()
                return

            self._flag_modified_externally = False

        if not self.bibolamaziFile:
            logger.debug("No file loaded.")
            return

        self._set_modified(True)
        logger.debug('modified!!')

        self.bibolamaziFile.setConfigData(str(self.ui.txtConfig.toPlainText()))

        # set a short time-out to reparse the file. This is to avoid reparsing
        # too often, especially when doing
        # computationally/transactionally-intensive changes such as changing the
        # argument of a 'package:' directive with a remote location.
        self._delayed_reparse_textedit()


    @pyqtSlot()
    def _delayed_reparse_textedit(self):
        if self.reparseTimer.isActive():
            self.reparseTimer.stop()

        self.reparseTimer.setInterval(500)
        self.reparseTimer.setSingleShot(True)
        self.reparseTimer.start()

    @pyqtSlot()
    def _reparse_textedit(self):
        self._bibolamazifile_reparse()
        self._do_update_edittools()


    def _bibolamazifile_reparse(self):
        try:
            self.bibolamaziFile.load(to_state=bibolamazifile.BIBOLAMAZIFILE_PARSED)
        except BibolamaziError as e:
            # see if we can parse the error
            self.ui.txtParseErrorMessages.setVisible(True)
            self.ui.txtParseErrorMessages.setHtml(
                "<p style=\"color: rgb(127,0,0)\">Parse Error in file:</p>\n"+
                bibolamazi_error_html(str(e)))
            # self.ui.txtInfo.setHtml("<p style=\"color: rgb(180,0,0)\">Parse error in file&mdash;"
            #                         "check \"Run messages\" tab.</p>")
            return
        self.ui.txtParseErrorMessages.setVisible(False)

    @pyqtSlot()
    def on_txtConfig_cursorPositionChanged(self):
        logger.debug("cursor position changed!")
        
        self._do_update_edittools()


    @pyqtSlot()
    def on_btnAddFilter_clicked(self):
        logger.debug('add filter: clicked')

        filterpath = filterfactory.filterpath
        if self.bibolamaziFile.getLoadState() == bibolamazifile.BIBOLAMAZIFILE_PARSED:
            filterpath = self.bibolamaziFile.fullFilterPath()

        filter_list = filterinstanceeditor.get_filter_list(filterpath)

        (filtname, ok) = QInputDialog.getItem(self, "Select Filter",
                                              "Please select which filter you wish to add",
                                              filter_list)
        if not ok:
            # cancelled
            return

        # insert new filter command
        self._insert_new_cmd('filter: %s' %(filtname))

    @pyqtSlot()
    def on_btnAddSourceList_clicked(self):
        logger.debug('add source list: clicked')

        self._insert_new_cmd('src: ""')
        # directly set the focus on the file name field
        self.ui.sourceListEditor.changeSource()
        

    def _insert_new_cmd(self, cmdtext):

        doc = self.ui.txtConfig.document()
        
        insertcur = None
        
        cmd = self._get_current_bibolamazi_cmd()
        if cmd is None:
            insertcur = QTextCursor(doc.findBlockByNumber(self.ui.txtConfig.textCursor().block().blockNumber()))
        else:
            # insert _after_ current cmd (-> +1 for next line, -1 to correct for offset starting at 1)
            insertcur = QTextCursor(doc.findBlockByNumber(self.bibolamaziFile.configLineNo(cmd.linenoend)))

        cmdtext = str(cmdtext)

        insertcur.insertText('\n' + cmdtext+'\n\n')
        # select inserted text without the newline
        #insertcur.movePosition(QTextCursor.Left)
        insertcur.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, len(cmdtext)+1)
        #insertcur.movePosition(QTextCursor.StartOfBlock, QTextCursor.KeepAnchor)
        self.ui.txtConfig.setTextCursor(insertcur)


    def _get_current_bibolamazi_cmd(self):

        if not self.bibolamaziFile:
            logger.warning("No bibolamazi file open")
            return

        cmds = self.bibolamaziFile.configCmds()
        if cmds is None:
            # file is not yet parsed
            return None

        cur = self.ui.txtConfig.textCursor()
        block = cur.block()
        thisline = cur.block().blockNumber()+1

        thisline = self.bibolamaziFile.fileLineNo(thisline)

        logger.longdebug("searching for cmd... at file line=%d", thisline)
        for cmd in cmds:
            logger.longdebug("\t -- testing cmd %r", cmd)
            if cmd.lineno <= thisline and cmd.linenoend >= thisline:
                # got the current cmd
                logger.longdebug("\tGot cmd: %r", cmd)
                return cmd

        return None

    def _get_current_bibolamazi_cmdlist(self):

        if not self.bibolamaziFile:
            logger.warning("No bibolamazi file open")
            return

        cmds = self.bibolamaziFile.configCmds()
        if cmds is None:
            # file is not yet parsed
            return None

        cur = self.ui.txtConfig.textCursor()
        block = cur.block()
        thisline = cur.block().blockNumber()+1
        thisline = self.bibolamaziFile.fileLineNo(thisline)

        if cur.hasSelection():
            cur2 = QTextCursor(cur)
            cur2.setPosition(cur.anchor())
            otherline = cur2.block().blockNumber()+1
            otherline = self.bibolamaziFile.fileLineNo(otherline)
        else:
            otherline = thisline

        minline = min(thisline, otherline)
        maxline = max(thisline, otherline)

        logger.longdebug("searching for cmds between lines %d and %d", minline, maxline)
        cmdlist = []
        for cmd in cmds:
            logger.longdebug("\t -- testing cmd %r", cmd)
            if cmd.lineno <= maxline and cmd.linenoend >= minline:
                # got the current cmd
                logger.longdebug("\tGot cmd: %r", cmd)
                cmdlist.append(cmd)

        return cmdlist


    def _do_update_edittools(self):
        if self._ignore_change_for_edittools:
            return

        cmdlist = self._get_current_bibolamazi_cmdlist()
        if not cmdlist: # None, or empty list
            self.ui.stackEditTools.setCurrentWidget(self.ui.toolspageBase)
            return

        if len(cmdlist) > 1:
            # several commands selected
            self.ui.stackEditTools.setCurrentWidget(self.ui.toolspageSelectedMultiple)
            return

        # single command selected
        cmd = cmdlist[0]
        
        if cmd.cmd == "src":
            try:
                thesrcs = shlex.split(cmd.text)
            except ValueError as e:
                logger.debug("Invalid source list specification, syntax error: %s", str(e))
                self.ui.sourceListEditor.dispSourceListError(str(e))
                return
            self.ui.sourceListEditor.setSourceList(thesrcs, noemit=True)
            self.ui.sourceListEditor.setRefDir(self.bibolamaziFile.fdir())
            self.ui.stackEditTools.setCurrentWidget(self.ui.toolspageSource)
            return

        if cmd.cmd == "package":
            try:
                self.ui.filterPackagePathEditor.setFilterPackageInfo(cmd.text.strip())
                self.ui.filterPackagePathEditor.setRefDir(self.bibolamaziFile.fdir())
            except BibolamaziError as e:
                self.ui.filterPackagePathEditor.setFilterPackageError(str(e))
            self.ui.stackEditTools.setCurrentWidget(self.ui.toolspagePackage)
            return

        if cmd.cmd == "filter":
            filtername = cmd.info['filtername']
            text = cmd.text
            if text and text[-1] == '\n':
                text = text[:-1]
            self.ui.filterInstanceEditor.setFilterPath(self.bibolamaziFile.fullFilterPath())
            self.ui.filterInstanceEditor.setFilterInstanceDefinition(filtername, text, noemit=True)
            self.ui.stackEditTools.setCurrentWidget(self.ui.toolspageFilter)
            return

        logger.warning("Unknown command: %r", cmd)
        return


    def _replace_current_cmd(self, repltext, forcecheckcmd):
        logger.debug('_replace_current_cmd(%r,%r)', repltext, forcecheckcmd)
        
        cmd = self._get_current_bibolamazi_cmd()

        if cmd is None or cmd.cmd != forcecheckcmd:
            logger.debug("Expected to currently be in cmd %s!!", forcecheckcmd)
            return

        logger.debug('About to change cmd %s on lines=%d--%d', cmd.cmd, cmd.lineno, cmd.linenoend)

        configlineno = self.bibolamaziFile.configLineNo(cmd.lineno)
        configlinenoend = self.bibolamaziFile.configLineNo(cmd.linenoend)

        self._ignore_change_for_edittools = True
        doc = self.ui.txtConfig.document()
        cursor = QTextCursor(doc.findBlockByNumber(configlineno-1))
        endblock = (configlinenoend+1)-1
        if endblock >= doc.blockCount():
            cursor.movePosition(QTextCursor.End, QTextCursor.KeepAnchor)
        else:
            cursorend = QTextCursor(doc.findBlockByNumber(endblock))
            cursor.setPosition(cursorend.position(), QTextCursor.KeepAnchor)
        cursor.insertText(repltext)
        tcursor = QTextCursor(doc.findBlockByNumber(configlineno-1))
        self.ui.txtConfig.setTextCursor(tcursor)
        self._ignore_change_for_edittools = False

        # now, reparse the config
        self.bibolamaziFile.setConfigData(str(self.ui.txtConfig.toPlainText()))
        self._bibolamazifile_reparse()
        

    @pyqtSlot('QStringList')
    def on_sourceListEditor_sourceListChanged(self, sourcelist):

        sourcelist = [str(x) for x in list(sourcelist)]

        cmdtext = "src: " + ("\n     ".join([butils.quotearg(x) for x in sourcelist])) + "\n"

        self._replace_current_cmd(cmdtext, "src")


    @pyqtSlot(str)
    def on_filterPackagePathEditor_filterPackagePathChanged(self, fpath):

        self._replace_current_cmd("package: " + fpath + "\n", "package")


    @pyqtSlot()
    def on_filterInstanceEditor_filterInstanceDefinitionChanged(self):
        filtername = self.ui.filterInstanceEditor.filterName()
        optionstring = self.ui.filterInstanceEditor.optionString()
        errorstring = self.ui.filterInstanceEditor.errorString()

        logger.debug('on_filterInstanceEditor_filterInstanceDefinitionChanged() '
                     'filtername=%r, optionstring=%r, errorstring=%r',
                     filtername, optionstring, errorstring)

        if errorstring is not None:
            return

        cmdtext = "filter: " + filtername + ' ' + optionstring + "\n"

        self._replace_current_cmd(cmdtext, "filter")
        

    @pyqtSlot('QString')
    def on_filterInstanceEditor_filterHelpRequested(self, topic):
        self.requestHelpTopic.emit(str(topic))


    @pyqtSlot()
    def add_favorite_cmd(self):
        cmd = self._get_current_bibolamazi_cmd()

        if cmd is None:
            logger.warning("No command to add to favorites!")
            return

        logger.debug("Adding command %s on lines %d--%d to favorites: %r",
                     cmd.cmd, cmd.lineno, cmd.linenoend, cmd)

        cmdtext = []
        doc = self.ui.txtConfig.document()
        for n in range(self.bibolamaziFile.configLineNo(cmd.lineno),
                       self.bibolamaziFile.configLineNo(cmd.linenoend)+1):
            cmdtext.append(str(doc.findBlockByNumber(n-1).text()))
        cmdtext = "\n".join(cmdtext)

        self.add_favorite_cmdtext(cmdtext)

    @pyqtSlot()
    def add_favorite_selection(self):
        selected_text = self.ui.txtConfig.textCursor().selectedText()

        logger.debug("Selected text: %r", selected_text)
        self.add_favorite_cmdtext(selected_text)

    @pyqtSlot(str)
    def add_favorite_cmdtext(self, cmdtext):
        cmdtext = str(cmdtext)

        shorttext = cmdtext
        if len(shorttext) > 27:
            shorttext = shorttext[:30] + '...'

        self.favoriteCmdsList.addFavorite(FavoriteCmd(name=shorttext, cmd=cmdtext))

        QMessageBox.information(self, "Favorite Added", "Added this command to your favorites.")



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'newbibolamazifiledialog.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_NewBibolamazifileDialog(object):
    def setupUi(self, NewBibolamazifileDialog):
        NewBibolamazifileDialog.setObjectName("NewBibolamazifileDialog")
        NewBibolamazifileDialog.resize(528, 592)
        NewBibolamazifileDialog.setSizeGripEnabled(True)
        self.verticalLayout = QtWidgets.QVBoxLayout(NewBibolamazifileDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(NewBibolamazifileDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setMinimumSize(QtCore.QSize(0, 30))
        font = QtGui.QFont()
        font.setItalic(True)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.stk = QtWidgets.QStackedWidget(NewBibolamazifileDialog)
        self.stk.setObjectName("stk")
        self.page = QtWidgets.QWidget()
        self.page.setObjectName("page")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.page)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.srcmulti = QtWidgets.QGroupBox(self.page)
        self.srcmulti.setObjectName("srcmulti")
        self.gridLayout = QtWidgets.QGridLayout(self.srcmulti)
        self.gridLayout.setObjectName("gridLayout")
        self.rdbtnMergeMultiple = QtWidgets.QRadioButton(self.srcmulti)
        self.rdbtnMergeMultiple.setObjectName("rdbtnMergeMultiple")
        self.gridLayout.addWidget(self.rdbtnMergeMultiple, 1, 0, 1, 2)
        self.rdbtnSingleSource = QtWidgets.QRadioButton(self.srcmulti)
        self.rdbtnSingleSource.setChecked(True)
        self.rdbtnSingleSource.setObjectName("rdbtnSingleSource")
        self.gridLayout.addWidget(self.rdbtnSingleSource, 0, 0, 1, 2)
        self.verticalLayout_2.addWidget(self.srcmulti)
        self.groupBox_2 = QtWidgets.QGroupBox(self.page)
        self.groupBox_2.setObjectName("groupBox_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.groupBox_2)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.txtSources = QtWidgets.QTextBrowser(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.txtSources.sizePolicy().hasHeightForWidth())
        self.txtSources.setSizePolicy(sizePolicy)
        self.txtSources.setFrameShape(QtWidgets.QFrame.Panel)
        self.txtSources.setTabChangesFocus(True)
        self.txtSources.setObjectName("txtSources")
        self.horizontalLayout.addWidget(self.txtSources)
        self.verticalLayout_9 = QtWidgets.QVBoxLayout()
        self.verticalLayout_9.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.btnSrcSet = QtWidgets.QPushButton(self.groupBox_2)
        self.btnSrcSet.setObjectName("btnSrcSet")
        self.verticalLayout_9.addWidget(self.btnSrcSet)
        self.btnSrcAdd = QtWidgets.QPushButton(self.groupBox_2)
        self.btnSrcAdd.setObjectName("btnSrcAdd")
        self.verticalLayout_9.addWidget(self.btnSrcAdd)
        self.btnSrcClear = QtWidgets.QPushButton(self.groupBox_2)
        self.btnSrcClear.setObjectName("btnSrcClear")
        self.verticalLayout_9.addWidget(self.btnSrcClear)
        self.horizontalLayout.addLayout(self.verticalLayout_9)
        self.verticalLayout_2.addWidget(self.groupBox_2)
        self.chkDuplicatesFilter = QtWidgets.QGroupBox(self.page)
        self.chkDuplicatesFilter.setCheckable(True)
        self.chkDuplicatesFilter.setObjectName("chkDuplicatesFilter")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.chkDuplicatesFilter)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label_5 = QtWidgets.QLabel(self.chkDuplicatesFilter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_5.sizePolicy().hasHeightForWidth())
        self.label_5.setSizePolicy(sizePolicy)
        self.label_5.setWordWrap(True)
        self.label_5.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByKeyboard|QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextBrowserInteraction|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.label_5.setObjectName("label_5")
        self.verticalLayout_5.addWidget(self.label_5)
        self.verticalLayout_2.addWidget(self.chkDuplicatesFilter)
        spacerItem = QtWidgets.QSpacerItem(20, 5, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.stk.addWidget(self.page)
        self.page_2 = QtWidgets.QWidget()
        self.page_2.setObjectName("page_2")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.page_2)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.chkArxivUnpublished = QtWidgets.QGroupBox(self.page_2)
        self.chkArxivUnpublished.setCheckable(True)
        self.chkArxivUnpublished.setObjectName("chkArxivUnpublished")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.chkArxivUnpublished)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.rdbtnArxivUnpub_unpubnote = QtWidgets.QRadioButton(self.chkArxivUnpublished)
        self.rdbtnArxivUnpub_unpubnote.setObjectName("rdbtnArxivUnpub_unpubnote")
        self.gridLayout_2.addWidget(self.rdbtnArxivUnpub_unpubnote, 2, 0, 1, 1)
        self.rdbtnArxivUnpub_eprint = QtWidgets.QRadioButton(self.chkArxivUnpublished)
        self.rdbtnArxivUnpub_eprint.setChecked(True)
        self.rdbtnArxivUnpub_eprint.setObjectName("rdbtnArxivUnpub_eprint")
        self.gridLayout_2.addWidget(self.rdbtnArxivUnpub_eprint, 1, 0, 1, 1)
        self.chkArxivUnpubIncludeTheses = QtWidgets.QCheckBox(self.chkArxivUnpublished)
        self.chkArxivUnpubIncludeTheses.setChecked(True)
        self.chkArxivUnpubIncludeTheses.setObjectName("chkArxivUnpubIncludeTheses")
        self.gridLayout_2.addWidget(self.chkArxivUnpubIncludeTheses, 3, 0, 1, 1)
        self.verticalLayout_3.addWidget(self.chkArxivUnpublished)
        self.chkArxivPublished = QtWidgets.QGroupBox(self.page_2)
        self.chkArxivPublished.setCheckable(True)
        self.chkArxivPublished.setObjectName("chkArxivPublished")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.chkArxivPublished)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.rdbtnArxivPub_eprint = QtWidgets.QRadioButton(self.chkArxivPublished)
        self.rdbtnArxivPub_eprint.setChecked(True)
        self.rdbtnArxivPub_eprint.setObjectName("rdbtnArxivPub_eprint")
        self.verticalLayout_7.addWidget(self.rdbtnArxivPub_eprint)
        self.rdbtnArxivPub_unpubnote = QtWidgets.QRadioButton(self.chkArxivPublished)
        self.rdbtnArxivPub_unpubnote.setObjectName("rdbtnArxivPub_unpubnote")
        self.verticalLayout_7.addWidget(self.rdbtnArxivPub_unpubnote)
        self.rdbtnArxivPub_strip = QtWidgets.QRadioButton(self.chkArxivPublished)
        self.rdbtnArxivPub_strip.setObjectName("rdbtnArxivPub_strip")
        self.verticalLayout_7.addWidget(self.rdbtnArxivPub_strip)
        self.chkArxivPubIncludeTheses = QtWidgets.QCheckBox(self.chkArxivPublished)
        self.chkArxivPubIncludeTheses.setObjectName("chkArxivPubIncludeTheses")
        self.verticalLayout_7.addWidget(self.chkArxivPubIncludeTheses)
        self.verticalLayout_3.addWidget(self.chkArxivPublished)
        self.chkUrlFilter = QtWidgets.QGroupBox(self.page_2)
        self.chkUrlFilter.setCheckable(True)
        self.chkUrlFilter.setObjectName("chkUrlFilter")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.chkUrlFilter)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.chkUrlStrip = QtWidgets.QCheckBox(self.chkUrlFilter)
        self.chkUrlStrip.setObjectName("chkUrlStrip")
        self.verticalLayout_8.addWidget(self.chkUrlStrip)
        self.chkUrlStripAllIfDoiOrArxiv = QtWidgets.QCheckBox(self.chkUrlFilter)
        self.chkUrlStripAllIfDoiOrArxiv.setChecked(True)
        self.chkUrlStripAllIfDoiOrArxiv.setObjectName("chkUrlStripAllIfDoiOrArxiv")
        self.verticalLayout_8.addWidget(self.chkUrlStripAllIfDoiOrArxiv)
        self.chkUrlStripDoiArxiv = QtWidgets.QCheckBox(self.chkUrlFilter)
        self.chkUrlStripDoiArxiv.setChecked(True)
        self.chkUrlStripDoiArxiv.setObjectName("chkUrlStripDoiArxiv")
        self.verticalLayout_8.addWidget(self.chkUrlStripDoiArxiv)
        self.chkUrlKeepFirstUrlOnly = QtWidgets.QCheckBox(self.chkUrlFilter)
        self.chkUrlKeepFirstUrlOnly.setChecked(True)
        self.chkUrlKeepFirstUrlOnly.setObjectName("chkUrlKeepFirstUrlOnly")
        self.verticalLayout_8.addWidget(self.chkUrlKeepFirstUrlOnly)
        self.verticalLayout_3.addWidget(self.chkUrlFilter)
        spacerItem1 = QtWidgets.QSpacerItem(20, 5, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem1)
        self.stk.addWidget(self.page_2)
        self.page_3 = QtWidgets.QWidget()
        self.page_3.setObjectName("page_3")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.page_3)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.chkFixesFilter = QtWidgets.QGroupBox(self.page_3)
        self.chkFixesFilter.setCheckable(True)
        self.chkFixesFilter.setObjectName("chkFixesFilter")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.chkFixesFilter)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.chkFixesGeneralSelection = QtWidgets.QCheckBox(self.chkFixesFilter)
        self.chkFixesGeneralSelection.setChecked(True)
        self.chkFixesGeneralSelection.setObjectName("chkFixesGeneralSelection")
        self.verticalLayout_6.addWidget(self.chkFixesGeneralSelection)
        self.chkFixesEncodeToLatex = QtWidgets.QCheckBox(self.chkFixesFilter)
        self.chkFixesEncodeToLatex.setChecked(True)
        self.chkFixesEncodeToLatex.setObjectName("chkFixesEncodeToLatex")
        self.verticalLayout_6.addWidget(self.chkFixesEncodeToLatex)
        self.verticalLayout_4.addWidget(self.chkFixesFilter)
        self.chkKeepOnlyUsed = QtWidgets.QGroupBox(self.page_3)
        self.chkKeepOnlyUsed.setCheckable(True)
        self.chkKeepOnlyUsed.setObjectName("chkKeepOnlyUsed")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.chkKeepOnlyUsed)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.label_4 = QtWidgets.QLabel(self.chkKeepOnlyUsed)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_4.sizePolicy().hasHeightForWidth())
        self.label_4.setSizePolicy(sizePolicy)
        self.label_4.setWordWrap(True)
        self.label_4.setObjectName("label_4")
        self.gridLayout_3.addWidget(self.label_4, 0, 0, 1, 1)
        self.verticalLayout_4.addWidget(self.chkKeepOnlyUsed)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_4.addItem(spacerItem2)
        self.stk.addWidget(self.page_3)
        self.verticalLayout.addWidget(self.stk)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.btnCancel = QtWidgets.QPushButton(NewBibolamazifileDialog)
        self.btnCancel.setObjectName("btnCancel")
        self.horizontalLayout_3.addWidget(self.btnCancel)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem3)
        self.btnBack = QtWidgets.QPushButton(NewBibolamazifileDialog)
        self.btnBack.setObjectName("btnBack")
        self.horizontalLayout_3.addWidget(self.btnBack)
        self.btnNext = QtWidgets.QPushButton(NewBibolamazifileDialog)
        self.btnNext.setObjectName("btnNext")
        self.horizontalLayout_3.addWidget(self.btnNext)
        self.btnSaveFinish = QtWidgets.QPushButton(NewBibolamazifileDialog)
        self.btnSaveFinish.setObjectName("btnSaveFinish")
        self.horizontalLayout_3.addWidget(self.btnSaveFinish)
        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.retranslateUi(NewBibolamazifileDialog)
        self.stk.setCurrentIndex(0)
        self.rdbtnSingleSource.toggled['bool'].connect(self.btnSrcSet.setVisible)
        self.rdbtnMergeMultiple.toggled['bool'].connect(self.btnSrcAdd.setVisible)
        self.rdbtnMergeMultiple.toggled['bool'].connect(self.btnSrcClear.setVisible)
        self.rdbtnMergeMultiple.toggled['bool'].connect(self.chkDuplicatesFilter.setVisible)
        self.btnCancel.clicked.connect(NewBibolamazifileDialog.reject)
        self.chkUrlStrip.toggled['bool'].connect(self.chkUrlStripAllIfDoiOrArxiv.setDisabled)
        self.chkUrlStrip.toggled['bool'].connect(self.chkUrlKeepFirstUrlOnly.setDisabled)
        self.chkUrlStrip.toggled['bool'].connect(self.chkUrlStripDoiArxiv.setDisabled)
        QtCore.QMetaObject.connectSlotsByName(NewBibolamazifileDialog)

    def retranslateUi(self, NewBibolamazifileDialog):
        _translate = QtCore.QCoreApplication.translate
        NewBibolamazifileDialog.setWindowTitle(_translate("NewBibolamazifileDialog", "Create new bibolamazi file"))
        self.label.setText(_translate("NewBibolamazifileDialog", "So you\'d like to organize some bibtex entries..."))
        self.srcmulti.setTitle(_translate("NewBibolamazifileDialog", "Would you like to collect entries from several bibtex files?"))
        self.rdbtnMergeMultiple.setText(_translate("NewBibolamazifileDialog", "I would like to merge bibliography entries from different files"))
        self.rdbtnSingleSource.setText(_translate("NewBibolamazifileDialog", "My bibliography entries are currently all in one bibtex file"))
        self.groupBox_2.setTitle(_translate("NewBibolamazifileDialog", "Specify sources where to collect bibtex entries from"))
        self.txtSources.setHtml(_translate("NewBibolamazifileDialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'.SF NS Text\'; font-size:13pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>"))
        self.btnSrcSet.setText(_translate("NewBibolamazifileDialog", "set"))
        self.btnSrcAdd.setText(_translate("NewBibolamazifileDialog", "add"))
        self.btnSrcClear.setText(_translate("NewBibolamazifileDialog", "clear"))
        self.chkDuplicatesFilter.setTitle(_translate("NewBibolamazifileDialog", "Merge duplicate entries with different key names && create aliases"))
        self.label_5.setText(_translate("NewBibolamazifileDialog", "<html><head/><body>\n"
"<p>Citing the same article using different keys, such as <code>\\cite{Einstein1937}</code> and <code>\\cite{EPRpaper}</code> will refer to the same bibliography entry.</p>\n"
"<p style=\"font-style:italic\">In order for this to work, you need to add the following line to the preamble of your document:</p><pre>\\input{bibolamazi_dup_aliases.tex}</pre></body></html>"))
        self.chkArxivUnpublished.setTitle(_translate("NewBibolamazifileDialog", "Normalize arxiv fields in entries, for *unpublished* entries"))
        self.rdbtnArxivUnpub_unpubnote.setText(_translate("NewBibolamazifileDialog", "Set entry type to @unpublished and add \"note = {arXiv:XXXX.ZZZZZ}\""))
        self.rdbtnArxivUnpub_eprint.setText(_translate("NewBibolamazifileDialog", "Use \"eprint\" field, like \"eprint = {XXXX.ZZZZZ}\""))
        self.chkArxivUnpubIncludeTheses.setText(_translate("NewBibolamazifileDialog", "Include theses as unpublished entries"))
        self.chkArxivPublished.setTitle(_translate("NewBibolamazifileDialog", "Normalize arxiv fields in entries, for *published* entries"))
        self.rdbtnArxivPub_eprint.setText(_translate("NewBibolamazifileDialog", "Use \"eprint\" field, like \"eprint = {XXXX.ZZZZZ}\""))
        self.rdbtnArxivPub_unpubnote.setText(_translate("NewBibolamazifileDialog", "Add \"note = {arXiv:XXXX.ZZZZZ}\""))
        self.rdbtnArxivPub_strip.setText(_translate("NewBibolamazifileDialog", "Strip arxiv information entirely from published entries"))
        self.chkArxivPubIncludeTheses.setText(_translate("NewBibolamazifileDialog", "Include theses as published entries"))
        self.chkUrlFilter.setTitle(_translate("NewBibolamazifileDialog", "Fix URLs"))
        self.chkUrlStrip.setText(_translate("NewBibolamazifileDialog", "Remove all URLs"))
        self.chkUrlStripAllIfDoiOrArxiv.setText(_translate("NewBibolamazifileDialog", "Remove all URLs if there is an arXiv id or a DOI"))
        self.chkUrlStripDoiArxiv.setText(_translate("NewBibolamazifileDialog", "Remove URLs that simply refer to a DOI or to an arxiv page"))
        self.chkUrlKeepFirstUrlOnly.setText(_translate("NewBibolamazifileDialog", "Keep only the first URL available"))
        self.chkFixesFilter.setTitle(_translate("NewBibolamazifileDialog", "Apply general fixes"))
        self.chkFixesGeneralSelection.setText(_translate("NewBibolamazifileDialog", "Include a selection of general fixes"))
        self.chkFixesEncodeToLatex.setText(_translate("NewBibolamazifileDialog", "Encode accents and special characters into LaTeX"))
        self.chkKeepOnlyUsed.setTitle(_translate("NewBibolamazifileDialog", "Keep only entries that I actually use in my LaTeX document"))
        self.label_4.setText(_translate("NewBibolamazifileDialog", "<html><body><p style=\"font-style:italic\">Make sure when you save the bibolamazi file to give it the same base name as your LaTeX document. For instance, if your document is named <code>example.tex</code>, save the bibolamazifile as <code>example.bibolamazi.bib</code>.</p></body></html>"))
        self.btnCancel.setText(_translate("NewBibolamazifileDialog", "cancel"))
        self.btnBack.setText(_translate("NewBibolamazifileDialog", "◀︎ back"))
        self.btnNext.setText(_translate("NewBibolamazifileDialog", "next ▶"))
        self.btnSaveFinish.setText(_translate("NewBibolamazifileDialog", "save file ▶"))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################


# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr


import sys
import os
import os.path
import re
import logging
import subprocess
import datetime

logger = logging.getLogger(__name__)

import bibolamazi.init

from bibolamazi.core import bibolamazifile
from bibolamazi.core import main
from bibolamazi.core.butils import BibolamaziError
from bibolamazi.core.bibfilter import factory as filters_factory
from bibolamazi.core.bibfilter import argtypes
from bibolamazi.core import version as bibolamaziversion



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

#print("loaded Qt. executable=", sys.executable, ' meipass=', sys._MEIPASS)

from . import uiutils
from . import openbibfile
from . import helpbrowser
from . import settingswidget

from .favorites import FavoriteCmdsList
from .newbibolamazifiledialog import NewBibolamazifileDialog

from .qtauto.ui_startupwidget import Ui_StartupWidget


class StartupWidget(QWidget):
    def __init__(self, bibapp):
        super(StartupWidget, self).__init__()
        self.setWindowFlags(self.windowFlags() | Qt.Dialog)

        self.bibapp = bibapp

        self.ui = Ui_StartupWidget()
        self.ui.setupUi(self)

        # set up nice vector graphics on retina displays
        if sys.platform.startswith("darwin"):
            # use high-res SVG for retina displays
            retinaresolution = find_retina_resolution()
            if retinaresolution is not None:
                mydesktop = QApplication.desktop()
                # seems that myratio is not reliable (I get 1.77777..), so just use 2x
                #myratio = float(retinaresolution[0]) / mydesktop.width()
                myratio = 2
                logger.debug('myratio=%r', myratio)
                #if myratio > 1.01:
                self.mypict = QPicture()
                mypaint = QPainter(self.mypict)
                if uiutils.is_dark_mode(self):
                    self.myicon = QIcon(":/pic/bibolamazi-darkmode.svg")
                else:
                    self.myicon = QIcon(":/pic/bibolamazi.svg")
                mysize = QSize(375, 150)
                mypaint.drawPixmap(QRect(QPoint(0,0),mysize), self.myicon.pixmap(myratio*mysize))
                self.ui.lblMain.setPicture(self.mypict)

        self.ui.btnOpenFile.clicked.connect(bibapp.selectOpenFile)
        self.ui.btnNewFile.clicked.connect(bibapp.createNewFile)

        self.recentfilesbtnmenu = QMenu(parent=self)
        self.bibapp.setRecentFilesMenu(self.recentfilesbtnmenu)
        self.bibapp.recentFilesAvailable.connect(self.ui.btnOpenRecent.setEnabled)
        self.ui.btnOpenRecent.setMenu(self.recentfilesbtnmenu)
        self.ui.btnOpenRecent.clicked.connect(self.ui.btnOpenRecent.showMenu)

        self.ui.btnSettings.clicked.connect(bibapp.openSettings)
        self.ui.btnHelp.clicked.connect(bibapp.openHelpBrowser)

        self.ui.btnQuit.clicked.connect(bibapp.quit_app)

        for a in bibapp.myactions.values():
            self.addAction(a)

        if not sys.platform.startswith('darwin'):
            # NOT mac OS X, no application menu bar.
            # Simply set up the shortcuts
            for a in bibapp.myactions.values():
                a.setShortcutContext(Qt.ApplicationShortcut)

        self.setWindowIcon(QIcon(':/pic/bibolamazi_icon.png'))

    def showAndRaise(self):
        logger.debug("showing & raising startup widget ...")
        self.show()
        self.raise_()
        logger.debug("showing & raising startup widget ... done.")

    def closeEvent(self, event):
        # Initiate an application quit if either:
        # - there are no files open any more OR 
        # - if we're not on a mac (and hence there is no app menu bar)
        if not self.bibapp.hasOpenFiles() or not sys.platform.startswith('darwin'):
            event.ignore()
            # get out of event loop first, though -> QTimer
            QTimer.singleShot(100, self.bibapp.quit_app)
            return

        super(StartupWidget, self).closeEvent(event)

#


def find_retina_resolution():
    try:
        output = subprocess.check_output(['system_profiler', 'SPDisplaysDataType']).decode('utf-8')
    except Exception as e:
        logger.info("Couldn't check for retina display: %s", e)
        return
    logger.debug("Got display information:\n%s", output)

    mres = re.search(r'Resolution:\s*(?P<resX>\d+)\s*x\s*(?P<resY>\d+)\s*(?P<retina>.*Retina)',
                     output, flags=re.IGNORECASE)

    if not mres.group('retina'):
        # couldn't confirm it's Retina, try "Retina: Yes/No" field
        mretina = re.search(r'Retina:\s*(?P<answer>Yes|No)', output, flags=re.IGNORECASE)
        if not mretina:
            return None

    # yes, retina
    logger.debug('got res=(%d,%d)', int(mres.group('resX')), int(mres.group('resY')))

    return (int(mres.group('resX')), int(mres.group('resY')))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'searchwidget.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_SearchWidget(object):
    def setupUi(self, SearchWidget):
        SearchWidget.setObjectName("SearchWidget")
        SearchWidget.resize(438, 28)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(SearchWidget.sizePolicy().hasHeightForWidth())
        SearchWidget.setSizePolicy(sizePolicy)
        self.horizontalLayout = QtWidgets.QHBoxLayout(SearchWidget)
        self.horizontalLayout.setContentsMargins(3, 3, 3, 3)
        self.horizontalLayout.setSpacing(2)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.lbl = QtWidgets.QLabel(SearchWidget)
        self.lbl.setObjectName("lbl")
        self.horizontalLayout.addWidget(self.lbl)
        self.txt = QtWidgets.QLineEdit(SearchWidget)
        self.txt.setObjectName("txt")
        self.horizontalLayout.addWidget(self.txt)
        self.btnNext = QtWidgets.QToolButton(SearchWidget)
        self.btnNext.setObjectName("btnNext")
        self.horizontalLayout.addWidget(self.btnNext)
        self.btnPrev = QtWidgets.QToolButton(SearchWidget)
        self.btnPrev.setObjectName("btnPrev")
        self.horizontalLayout.addWidget(self.btnPrev)
        self.btnDone = QtWidgets.QToolButton(SearchWidget)
        self.btnDone.setObjectName("btnDone")
        self.horizontalLayout.addWidget(self.btnDone)

        self.retranslateUi(SearchWidget)
        QtCore.QMetaObject.connectSlotsByName(SearchWidget)

    def retranslateUi(self, SearchWidget):
        _translate = QtCore.QCoreApplication.translate
        SearchWidget.setWindowTitle(_translate("SearchWidget", "Form"))
        self.lbl.setText(_translate("SearchWidget", "find: "))
        self.btnNext.setText(_translate("SearchWidget", "▼"))
        self.btnPrev.setText(_translate("SearchWidget", "▲"))
        self.btnDone.setText(_translate("SearchWidget", "done"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'favoritesoverbtns.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_FavoritesOverBtns(object):
    def setupUi(self, FavoritesOverBtns):
        FavoritesOverBtns.setObjectName("FavoritesOverBtns")
        FavoritesOverBtns.resize(197, 32)
        FavoritesOverBtns.setWindowOpacity(0.8)
        self.horizontalLayout = QtWidgets.QHBoxLayout(FavoritesOverBtns)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(5, 10, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.btnEndEdit = QtWidgets.QToolButton(FavoritesOverBtns)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/pic/ok.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnEndEdit.setIcon(icon)
        self.btnEndEdit.setIconSize(QtCore.QSize(10, 16))
        self.btnEndEdit.setObjectName("btnEndEdit")
        self.horizontalLayout.addWidget(self.btnEndEdit)
        self.btnEdit = QtWidgets.QToolButton(FavoritesOverBtns)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.btnEdit.sizePolicy().hasHeightForWidth())
        self.btnEdit.setSizePolicy(sizePolicy)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/pic/lstbtnedit.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnEdit.setIcon(icon1)
        self.btnEdit.setIconSize(QtCore.QSize(10, 16))
        self.btnEdit.setObjectName("btnEdit")
        self.horizontalLayout.addWidget(self.btnEdit)
        self.btnDelete = QtWidgets.QToolButton(FavoritesOverBtns)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/pic/lstbtnremove.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnDelete.setIcon(icon2)
        self.btnDelete.setIconSize(QtCore.QSize(10, 16))
        self.btnDelete.setObjectName("btnDelete")
        self.horizontalLayout.addWidget(self.btnDelete)
        spacerItem1 = QtWidgets.QSpacerItem(20, 5, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.btnInsert = QtWidgets.QPushButton(FavoritesOverBtns)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.btnInsert.sizePolicy().hasHeightForWidth())
        self.btnInsert.setSizePolicy(sizePolicy)
        self.btnInsert.setAutoFillBackground(True)
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/pic/rightarrow.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnInsert.setIcon(icon3)
        self.btnInsert.setIconSize(QtCore.QSize(30, 16))
        self.btnInsert.setFlat(True)
        self.btnInsert.setObjectName("btnInsert")
        self.horizontalLayout.addWidget(self.btnInsert)

        self.retranslateUi(FavoritesOverBtns)
        QtCore.QMetaObject.connectSlotsByName(FavoritesOverBtns)

    def retranslateUi(self, FavoritesOverBtns):
        _translate = QtCore.QCoreApplication.translate
        self.btnEndEdit.setToolTip(_translate("FavoritesOverBtns", "Finish editing"))
        self.btnEdit.setToolTip(_translate("FavoritesOverBtns", "Edit favorite commands"))
        self.btnDelete.setToolTip(_translate("FavoritesOverBtns", "Delete this favorite command"))
        self.btnInsert.setToolTip(_translate("FavoritesOverBtns", "Insert this favorite command at current position in config"))
        self.btnInsert.setText(_translate("FavoritesOverBtns", "insert"))

from . import bibolamazi_res_rc



################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2014 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################


# This is the current BIBOLAMAZI version.
# Bump the number here for different version numbers.

version_str = "4.2"
"""
The version string. This is increased upon each release.
"""


copyright_year = "2019"
"""
Year of copyright.
"""



# -*- coding: utf-8 -*-

################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2013 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr
from future.standard_library import install_aliases
install_aliases()

import os
import os.path
import re
import logging
import textwrap
from collections import OrderedDict
from urllib.parse import urlencode, parse_qs
from urllib.parse import quote_plus as urlquoteplus

import bibolamazi.init
from bibolamazi.core.bibfilter import factory as filters_factory
from bibolamazi.core.bibfilter.factory import NoSuchFilter, NoSuchFilterPackage, FilterError
from bibolamazi.core import butils
from bibolamazi.core.bibfilter.argtypes import EnumArgType, MultiTypeArgType, \
    CommaStrList, ColonCommaStrDict, StrEditableArgType
from bibolamazi.core.helppages import htmlescape, forcewrap_long_lines

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *


from .qtauto.ui_filterinstanceeditor import Ui_FilterInstanceEditor

from . import overlistbuttonwidget
from .multitypeseditor import MultiTypesEditorWidget

logger = logging.getLogger(__name__)



def is_class_and_is_sub_class(x, kl):
    # graceful simple "no" answer if issubclass throws TypeError
    try:
        return issubclass(x, kl)
    except Exception:
        return False



def get_filter_list(filterpath=filters_factory.filterpath):
    logger.debug("filterinstanceeditor.get_filter_list()")
    filter_pkg_list = filters_factory.detect_filter_package_listings(filterpath=filterpath)
    logger.debug("filterinstanceeditor.get_filter_list(): filter_pkg_list=%r", filter_pkg_list)
    filter_list = []
    for (fpkg, flist) in filter_pkg_list.items():
        if fpkg == 'bibolamazi.filters':
            # built-in, ignore the filter package prefix.
            filter_list += [ finfo.filtername for finfo in flist]
        else:
            filter_list += [ fpkg+':'+finfo.filtername for finfo in flist ]
    logger.debug("filterinstanceeditor.get_filter_list(): filter_list=%r", filter_list)
    return filter_list




class RegisteredArgInputType:
    def __init__(self, typ, val):
        self.typ = typ
        self.type_arg_input = typ.type_arg_input
        self.value = val

    def createWidget(self, parent, option=None):
        if isinstance(self.type_arg_input, EnumArgType):
            cbx = QComboBox(parent)
            for val in self.type_arg_input.listofvalues:
                cbx.addItem(val)
            return cbx

        if isinstance(self.type_arg_input, MultiTypeArgType):
            return MultiTypesEditorWidget(self.type_arg_input, parent)

        if isinstance(self.type_arg_input, StrEditableArgType):
            f = QItemEditorFactory()
            return f.createEditor(QVariant.String, parent)

        logger.debug("Unknown type: type_arg_input=%r", self.type_arg_input)
        return None

    def setEditorData(self, editor):
        logger.debug("setEditorData(), self.value=%r", self.value)

        if isinstance(self.type_arg_input, EnumArgType):
            for i in range(editor.count()):
                if (str(editor.itemText(i)) == self.value):
                    editor.setCurrentIndex(i)
                    return
            return

        if isinstance(self.type_arg_input, MultiTypeArgType):
            # self.value is the given value being edited (bool -dXXX or string -sXXX=...)
            if isinstance(self.value, self.typ): # a multi_type_class instance directly
                editor.setTypeValue(self.value.valuetype, self.value.value)
            else:
                editor.setValue(self.value)
            return

        if isinstance(self.type_arg_input, StrEditableArgType):
            editor.setText(str(self.value))

        logger.debug("Unknown type: type_arg_input=%r", self.type_arg_input)
        return None

    def valueOf(self, editor):
        if isinstance(self.type_arg_input, EnumArgType):
            val = str(editor.itemText(editor.currentIndex()))
            logger.debug("GOT VALUE: %r", val)
            return val

        if isinstance(self.type_arg_input, MultiTypeArgType):
            t, v = editor.getTypeValue()
            logger.debug("GOT VALUE: %r, %r", t, v)
            return self.typ(t, v)

        if isinstance(self.type_arg_input, StrEditableArgType):
            return self.typ(editor.text())

        logger.debug("Unknown type: type_arg_input=%r", self.type_arg_input)
        return None
    


class DefaultFilterOptionsModel(QAbstractTableModel):
    def __init__(self, filterpathobj, parent=None):
        super(DefaultFilterOptionsModel, self).__init__(parent)
        self._filterpathobj = filterpathobj
        self._filtername = None
        self._finfo = None
        self._fopts = None
        self._optionstring = None
        self._optionstring_error = None
        self._optionstring_softerror = None


    optionStringChanged = pyqtSignal(str)

    optionStringSetError = pyqtSignal(str)
    optionStringSetSoftError = pyqtSignal(str)
    optionStringClearError = pyqtSignal()


    def optionstring(self):
        return self._optionstring

    def errorString(self):
        return self._optionstring_error

    def softErrorString(self):
        return self._optionstring_softerror


    def filterInfo(self):
        return self._finfo


    @pyqtSlot(str)
    def setFilterName(self, filtername, force=False, noemit=False,
                      reset_optionstring=True):

        logger.debug("[Model]/setFilterName(%r)", filtername)
        logger.debug("path in filterpathobj is %r", self._filterpathobj)

        if filtername is None:
            # not initialized yet.
            return

        filtername = str(filtername)

        if (not force and self._filtername == filtername):
            return
        
        self._filtername = filtername

        self.beginResetModel()
        self._finfo = None
        self._fopts = None
        try:
            # remember: FilterInfo() may also raise NoSuchFilter
            self._finfo = filters_factory.FilterInfo(filtername, filterpath=self._filterpathobj)
            self._fopts = self._finfo.defaultFilterOptions()
        except Exception as e: #(NoSuchFilter,NoSuchFilterPackage,FilterError) as e:
            stre = str(e)
            logger.warning("Cannot inspect filter: %s", stre)
            import traceback
            logger.debug("%s", traceback.format_exc())
            self._optionstring_error = stre
            self.optionStringSetError.emit(stre)
            return
        finally:
            self.endResetModel()

        # self.optionStringClearError will be emitted, if appropriate after
        # successful parsing of option string

        if (reset_optionstring):
            self.setOptionString('', force=True, noemit=noemit)
        else:
            self.setOptionString(self._optionstring, force=True, noemit=noemit)

        #if (not noemit):
        #    # we shan't emit anything anyway for a filter name change.
        #    pass


    @pyqtSlot(str)
    def setOptionString(self, optionstring, force=False, noemit=False):

        if not self._fopts:
            logger.warning("Can't set option string because we can't manage "
                           "filter options for this filter")
            return None

        optionstring = str(optionstring);

        # don't reset source list if it's the same. In particular, don't emit the changed signal.
        if (not force and optionstring == self._optionstring and self._optionstring_error is None):
            return

        if (self._fopts is None):
            # no managed options
            return

        # parse the options
        try:
            optspec = self._fopts.parse_optionstring_to_optspec(optionstring)
        except Exception as e:
            logger.debug("error parsing option string.")
            errstr = str(e)
            if not errstr:
                errstr = "Unknown error"
            self._optionstring_error = errstr
            # emit regardless of noemit; noemit refers to content changes...
            self.optionStringSetError.emit(errstr)
            return

        if self._optionstring_error is not None:
            # reset any earlier error
            self._optionstring_error = None
            self.optionStringClearError.emit()

        # if the arguments are syntactically valid, but semantically invalid,
        # then display an error string but let the user edit the options anyway
        filter_opts_valid = True
        try:
            (xpargs, xkwargs) = self._finfo.parseOptionStringArgs(optionstring)
            self._finfo.validateOptionStringArgs(xpargs, xkwargs)
        except Exception as e:
            logger.debug("option string is invalid for filter, displaying error")
            errstr = str(e)
            if not errstr:
                errstr = "Invalid option string (unknown error)"
            self._optionstring_softerror = errstr
            # emit regardless of noemit; noemit refers to content changes...
            self.optionStringSetSoftError.emit(errstr)
            # ... but proceed here ...
            filter_opts_valid = False

        if filter_opts_valid and self._optionstring_softerror is not None:
            # reset any earlier soft error
            self._optionstring_softerror = None
            self.optionStringClearError.emit()

        self._optionstring = optionstring

        # optspec from above. Use user-input, not deduced ones with filter's
        # default values, as parseOptionStringArgs() etc. returns!!
        pargs = optspec['_args']
        kwargs = optspec['kwargs']
        if pargs is None:
            pargs = []

        self._pargs = pargs
        self._kwargs = kwargs

        self._emitLayoutChanged()


    @pyqtSlot(str)
    def removeArgument(self, argname):
        argname = str(argname)
        
        logger.debug('remove argument: %r', argname)
        
        if (argname in self._kwargs):

            logger.debug('really removing argument!')
            
            del self._kwargs[argname]
            
            row = self.findArgByName(argname)
            idx = self.index(row, 2)
            self._update_optionstring()
            self.dataChanged.emit(idx,idx)
            self._emitOptionStringChanged()


    def rowCount(self, parent):
        if parent.isValid():
            return 0
        if (self._fopts is None):
            return 0
        return len(self._fopts.filterOptions())

    def columnCount(self, parent):
        if parent.isValid():
            return 0
        return 2
    


    def arginfoForIndex(self, index):
        if not self._fopts:
            return None
        
        filteroptions = self._fopts.filterOptions()
        #col = index.column()
        #if (col < 0 or col >= 2):
        #    return None
        row = index.row()
        if (row < 0 or row >= len(filteroptions)):
            return None
        return filteroptions[row]

    def argdocForIndex(self, index):
        arg = self.arginfoForIndex(index)
        if arg is None:
            return None
        return arg.doc
        

    def data(self, index, role=Qt.DisplayRole):
        if (self._fopts is None):
            return None
        
        filteroptions = self._fopts.filterOptions()

        col = index.column()
        row = index.row()

        if (row < 0 or row >= len(filteroptions)):
            return None

        # the argument specification of the current row (_ArgDoc namedtuple instance)
        arg = filteroptions[row]

        if (col == 0):
            # argument name
            if (role == Qt.DisplayRole):
                return str(self._fopts.getSOptNameFromArg(filteroptions[row].argname))

            # tool-tip documentation
            if (role == Qt.ToolTipRole):
                return textwrap.fill(arg.doc, width=80)

            return None

        # the value of the argument of the current row.
        val = self._kwargs.get(arg.argname)
        
        if (col == 1):
            # argument value

            if (role == overlistbuttonwidget.ROLE_OVERBUTTON):
                if (val is not None):
                    return overlistbuttonwidget.OVERBUTTON_REMOVE
                return overlistbuttonwidget.OVERBUTTON_ADD

            if (role == overlistbuttonwidget.ROLE_ARGNAME):
                return str(arg.argname)
            
            if (role == Qt.DisplayRole):
                if (val is None):
                    return None
                return str(val)

            # request editing value of argument
            if (role == Qt.EditRole):
                if arg.argtypename is not None:
                    typ = butils.resolve_type(arg.argtypename, self._finfo.fmodule)
                else:
                    typ = str
                if hasattr(typ, 'type_arg_input'):
                    editval = RegisteredArgInputType(typ, val)
                elif is_class_and_is_sub_class(typ, str) and val is None:
                    editval = typ('')
                elif typ in [bool, int, float, str]:
                    try:
                        editval = typ(val) # resort to Qt's editor for these types.
                    except Exception:
                        # in case typ(val) generates an exception because val is invalid ...?
                        editval = typ()
                else:
                    editval = str(val) # always resort to string so that something can be edited
                return editval
            return None

        return None


    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if (orientation == Qt.Vertical):
            return None

        if (section == 0):
            if (role == Qt.DisplayRole):
                return str(u"Filter Option")
            return None

        if (section == 1):
            if (role == Qt.DisplayRole):
                return str(u"Value")
            return None

        return None


    def flags(self, index):
        col = index.column()
        row = index.row()

        if (self._fopts is None):
            return Qt.NoItemFlags

        if (col == 0):
            return Qt.ItemIsSelectable | Qt.ItemIsEnabled
        if (col == 1):
            return Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsEditable

        logger.debug("DefaultFilterOptionsModel.flags(): BAD COLUMN: %d", col)
        return 0
    

    def setData(self, index, value, role=Qt.EditRole):
        
        col = index.column()
        row = index.row()

        if (col != 1):
            return False

        if (self._fopts is None):
            return False

        if (role != Qt.EditRole):
            return False

        if self._fopts:
            filteroptions = self._fopts.filterOptions()
        else:
            filteroptions = []

        if (row < 0 or row >= len(filteroptions)):
            return False

        arg = filteroptions[row]

        logger.debug("Got value: %r", value)

        # validate type
        typ = None
        if (arg.argtypename is not None):
            typ = butils.resolve_type(arg.argtypename, self._finfo.fmodule)
        if (typ == None):
            typ = str
            
        value = typ(value)

        logger.debug("Got final value: %r ; typ=%r", value, typ)

        self._kwargs[arg.argname] = value

        self._update_optionstring()

        self.dataChanged.emit(index, index)

        self._emitOptionStringChanged()

        logger.debug("_kwargs is %r", self._kwargs)
        return True

    def findArgByName(self, argname):
        if not self._fopts:
            return None

        filteroptions = self._fopts.filterOptions()

        for row in range(len(filteroptions)):
            if (filteroptions[row].argname == argname):
                return row

        return None
    

    def _update_optionstring(self):
        if not self._fopts:
            return None

        slist = []
        for arg in self._pargs:
            slist.append(butils.quotearg(arg))
            
        # iterate fopts arguments to preserve that ordering
        done_args = []
        for arg in self._fopts.filterOptions():
            v0 = self._kwargs.get(arg.argname,None)

            v = v0
            if hasattr(v0, 'type_arg_input') and hasattr(v0.type_arg_input, 'option_val_repr'):
                v = v0.type_arg_input.option_val_repr(v0)

            logger.debug("formatting options, %r -> %r (-> %r)", arg.argname, v0, v)
            if v is not None:
                soptarg = self._fopts.getSOptNameFromArg(arg.argname)
                if (arg.argtypename == 'bool' or isinstance(v, bool)):
                    slist.append('-d'+soptarg+('' if v else '=False'))
                else:
                    slist.append('-s'+soptarg+'='+butils.quotearg(str(v)))
                    
            done_args.append(arg.argname)

        # and add any extra args
        for (k,v) in self._kwargs.items():
            if (k in done_args):
                continue
            slist.append('-s' + k + '=' + butils.quotearg(str(v)))

        optstring = " ".join(slist)
        if len(optstring) > 80:
            # long option string, so keep each option on a separate line
            indentstr = " "*(len("filter: ") + len(self._filtername) + 1)
            optstring = ("\n" + indentstr).join(slist)

        self._optionstring = optstring

        logger.debug("option string is now %r", self._optionstring)
            

    def _emitOptionStringChanged(self):
        self.optionStringChanged.emit(self._optionstring)

    def _emitLayoutChanged(self):
        self.layoutChanged.emit()



class DefaultFilterOptionsDelegate(QStyledItemDelegate):
    def __init__(self, parentView=None):
        super(DefaultFilterOptionsDelegate, self).__init__(parentView)
        self._view = parentView


    def sizeHint(self, option, index):
        # add vertical spacing
        return super(DefaultFilterOptionsDelegate, self).sizeHint(option, index) + QSize(0,6)

    def createEditor(self, parent, option, index):
        if (index.column() != 1):
            return super(DefaultFilterOptionsDelegate, self).createEditor(parent, option, index)
        
        data = index.data(Qt.EditRole)
        if (isinstance(data, RegisteredArgInputType)):
            # do our own processing with some custom widgets.
            rr = data
            w = rr.createWidget(parent, option)
            w.setProperty('_RegisteredArgInputType', data)
            return w
        
        return super(DefaultFilterOptionsDelegate, self).createEditor(parent, option, index)
    
    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)
    
    def setEditorData(self, editor, index):
        rr = editor.property('_RegisteredArgInputType')
        if rr is None:
            return super(DefaultFilterOptionsDelegate, self).setEditorData(editor, index)
        
        rr.setEditorData(editor) # the value is already contained in the rr object

    def setModelData(self, editor, model, index):
        rr = editor.property('_RegisteredArgInputType')
        if rr is None:
            return super(DefaultFilterOptionsDelegate, self).setModelData(editor, model, index)
        
        val = rr.valueOf(editor)

        logger.debug("setModelData(): value is = %r", val)

        model.setData(index, val)




# ------------------------------------------------------------------------------



class FilterInstanceEditor(QWidget):
    def __init__(self, parent):
        super(FilterInstanceEditor, self).__init__(parent)

        self._is_updating = True

        self.ui = Ui_FilterInstanceEditor()
        self.ui.setupUi(self)

        self.filterpath = OrderedDict(filters_factory.filterpath)

        self.ui.btnAddFavorite.clicked.connect(self.requestAddToFavorites)

        self.filterNameChanged.connect(self.filterInstanceDefinitionChanged)
        self.filterOptionsChanged.connect(self.filterInstanceDefinitionChanged)

        self._filteroptionsmodel = DefaultFilterOptionsModel(filterpathobj=self.filterpath,
                                                             parent=self)

        self._filteroptionsdelegate = DefaultFilterOptionsDelegate(parentView=self.ui.lstOptions)
        
        self.ui.lstOptions.setModel(self._filteroptionsmodel)
        self.ui.lstOptions.setItemDelegate(self._filteroptionsdelegate)

        self.ui.lstOptions.selectionModel().currentChanged.connect(self.option_selected)

        self._filteroptionsmodel.optionStringChanged.connect(self.filterOptionsChanged)

        self._filterargbtn = overlistbuttonwidget.OverListButtonWidget(self.ui.lstOptions)
        self._filterargbtn.removeClicked.connect(self._filteroptionsmodel.removeArgument)
        self._filterargbtn.addIndexClicked.connect(self.ui.lstOptions.edit)
        self._filterargbtn.editIndexClicked.connect(self.ui.lstOptions.edit)

        self._filteroptionsmodel.optionStringChanged.connect(self._filterargbtn.updateDisplay)
        self._filteroptionsmodel.optionStringSetError.connect(self.show_optionstring_error)
        self._filteroptionsmodel.optionStringSetSoftError.connect(self.show_optionstring_softerror)
        self._filteroptionsmodel.optionStringClearError.connect(self.clear_optionstring_error)

        self._is_updating = False

        self.clear_optionstring_error()
        

    filterInstanceDefinitionChanged = pyqtSignal()
    filterNameChanged = pyqtSignal(str)
    filterOptionsChanged = pyqtSignal(str)

    filterHelpRequested = pyqtSignal(str)

    requestAddToFavorites = pyqtSignal()
    

    def filterName(self):
        return self.ui.lblFilterName.text()

    def optionString(self):
        return self._filteroptionsmodel.optionstring()

    def errorString(self):
        return self._filteroptionsmodel.errorString()

    def softErrorString(self):
        return self._filteroptionsmodel.softErrorString()

    @pyqtSlot(str, bool)
    @pyqtSlot(str)
    def setFilterName(self, filtername, noemit=False, force=False, reset_optionstring=True):

        logger.debug("setFilterName(%r)", filtername)
        
        if (not force and self.ui.lblFilterName.text() == filtername):
            return
        
        self._is_updating = True
        self.ui.lblFilterName.setText(filtername) #cbxFilter.setEditText(filtername)
        self._is_updating = False

        self._filteroptionsmodel.setFilterName(filtername, noemit=True,
                                               reset_optionstring=reset_optionstring)

        self.ui.lstOptions.resizeColumnToContents(0)
        
        if (not noemit):
            self.emitFilterNameChanged()

        logger.debug("setFilterName(%r) done.", filtername)


    @pyqtSlot(str)
    def setOptionString(self, optionstring, noemit=False, force=False):
        logger.debug("setOptionString(%r)", optionstring)
        self._filteroptionsmodel.setOptionString(optionstring, force=force, noemit=noemit)
        self.option_selected(self.ui.lstOptions.selectionModel().currentIndex())
        #self.ui.lstOptions.resizeColumnToContents(0)
        #self.ui.lstOptions.setColumnWidth(0, self.ui.lstOptions.width()/3)
        #self.ui.lstOptions.setColumnWidth(1, 100)
        ##self._filteroptionsdelegate.update_buttons()
        logger.debug("setOptionString() done")

    def setFilterPath(self, filterpath):
        # do not allocate new object for self.filterpath, it is referenced by
        # other objects that depend on it
        self.filterpath.clear()
        self.filterpath.update(filterpath)

    def setFilterInstanceDefinition(self, filtername, optionstring, noemit=False):
        logger.debug("setFilterInstanceDefinition(filtername=%r, optionstring=%r, noemit=%r)",
                     filtername, optionstring, noemit)
        self.setFilterName(filtername, noemit=noemit, force=True, reset_optionstring=True)
        self.setOptionString(optionstring, noemit=noemit)


    def _mk_errstr_disp(self, errstr):
        self.lbl_error_msg_raw_string = errstr
        if len(errstr) > 200:
            return htmlescape(errstr[:(200-8)])+'... <a href="errormsg:/showpopup">more</a>'
        else:
            return htmlescape(errstr)


    @pyqtSlot(str)
    def show_optionstring_error(self, errstr):
        self.ui.lblErrorMsg.setText("Error: " + self._mk_errstr_disp(errstr))
        self.ui.lblErrorMsg.setVisible(True)
        self.ui.lstOptions.setEnabled(False)

    @pyqtSlot(str)
    def show_optionstring_softerror(self, errstr):
        self.ui.lblErrorMsg.setText("Invalid options: " + self._mk_errstr_disp(errstr))
        self.ui.lblErrorMsg.setVisible(True)

    @pyqtSlot(str)
    def on_lblErrorMsg_linkActivated(self, link):
        if link == 'errormsg:/showpopup':
            QWhatsThis.showText(
                self.ui.lblErrorMsg.mapToGlobal(QPoint(0,0)),
                ('<pre style="color: rgb(127,0,0); background-color: #efefef; font-weight:bold">'+
                 htmlescape(forcewrap_long_lines(self.lbl_error_msg_raw_string))+'</pre>'),
                self
            )
        else:
            QMessageBox.warning(self, 'Unknown link activated', 'Unknown link activated: %s'%(link))


    @pyqtSlot()
    def clear_optionstring_error(self):
        self.ui.lblErrorMsg.setVisible(False)
        self.lbl_error_msg_raw_string = None
        self.ui.lstOptions.setEnabled(True)
        
    @pyqtSlot()
    def emitFilterNameChanged(self):
        if (self._is_updating):
            return
        logger.debug("emitting filterNameChanged! filterName=%s", self.filterName())
        self.filterNameChanged.emit(str(self.filterName()))


    def showEvent(self, event):

        self.ui.lstOptions.setColumnWidth(0, self.ui.lstOptions.width()/3)

        return super(FilterInstanceEditor, self).showEvent(event)

    
    @pyqtSlot('QModelIndex', 'QModelIndex')
    def option_selected(self, currentindex, previousindex=None):
        arg = self._filteroptionsmodel.arginfoForIndex(currentindex)
        if arg:
            doc = ("<p>" + htmlescape(arg.doc) +
                   " <a style=\"text-decoration:none\" href=\"action:/help?anchorname={}\">"\
                   .format(urlquoteplus('a-filter-option-'+arg.argname)) +
                   "more\N{HORIZONTAL ELLIPSIS}</a></p>")
            self.ui.lblOptionHelp.setText(doc)
            self.ui.lblOptionHelp.setVisible(True)
        else:
            self.ui.lblOptionHelp.setVisible(False)

    @pyqtSlot(str)
    def on_lblOptionHelp_linkActivated(self, link):
        link_help_m = re.match(r'action\:\/help(\?(?P<qs>.*))?', link)
        if link_help_m:
            qsstr = link_help_m.group('qs')
            qs = {}
            if qsstr:
                qs = parse_qs(qsstr)
            anchorname = "".join(qs.pop('anchorname', None)[-1:])
            self.request_filter_help(anchorname=anchorname, qs=qs)
        else:
            logger.warning("Invalid link action in filter instance: %r", link)

    @pyqtSlot()
    def on_btnFilterHelp_clicked(self):
        self.request_filter_help()

    def request_filter_help(self, anchorname=None, qs=None):
        finfo = self._filteroptionsmodel.filterInfo()

        if finfo is None:
            logger.warning("request_filter_help: No filter information available.")
            return
        
        url = 'help:/filter/%s' % ( str(finfo.filtername) )
        qs = dict(qs) if qs else {}
        qs.update(filterpackage=finfo.filterpackagespec)
        url += '?'+urlencode(qs)
        if anchorname:
            url += '#'+anchorname
        logger.debug("Filter help: requesting topic URL %s", url)
        self.filterHelpRequested.emit(url)

        



# -*- coding: utf-8 -*-
################################################################################
#                                                                              #
#   This file is part of the Bibolamazi Project.                               #
#   Copyright (C) 2018 by Philippe Faist                                       #
#   philippe.faist@bluewin.ch                                                  #
#                                                                              #
#   Bibolamazi is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   Bibolamazi is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with Bibolamazi.  If not, see <http://www.gnu.org/licenses/>.        #
#                                                                              #
################################################################################

# Py2/Py3 support
from __future__ import unicode_literals, print_function
from past.builtins import basestring
from future.utils import python_2_unicode_compatible, iteritems
from builtins import range
from builtins import str as unicodestr

import re
import logging
from html import escape as htmlescape

import github

import bibolamazi.init

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from .uiutils import ContextAttributeSetter, BlockedSignals
from .qtauto.ui_githubreposelector import Ui_GithubRepoSelector

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------


class GithubRepoSelector(QDialog):
    def __init__(self, parent):
        super(GithubRepoSelector, self).__init__(parent)

        self.ui = Ui_GithubRepoSelector()
        self.ui.setupUi(self)

        self.ui.stk.setCurrentWidget(self.ui.pageUsername)

        self.ui.cbxRepos.repolist = []

        self._update_gui_state()

    def getUsernameRepo(self):
        return (self.ui.txtUser.text(), self.ui.cbxRepos.currentText())

    def getFilterPackageUrl(self):
        return 'github:' + self.ui.txtUser.text() + '/' + self.ui.cbxRepos.currentText()

    @pyqtSlot(str)
    def on_txtUser_textChanged(self, txt):
        self._update_gui_state()

    @pyqtSlot()
    def on_btnBack_clicked(self):
        self.ui.stk.setCurrentWidget(self.ui.pageUsername)
        self._update_gui_state()

    @pyqtSlot()
    def on_btnNext_clicked(self):
        self.ui.stk.setCurrentWidget(self.ui.pageRepo)
        self._update_gui_state()

    @pyqtSlot(str)
    def on_cbxRepos_editTextChanged(self, text):
        self._update_gui_state()

    @pyqtSlot()
    def on_btnOk_clicked(self):
        self.accept()

    def _update_gui_state(self):
        if self.ui.stk.currentWidget() == self.ui.pageUsername:
            self.ui.stk.setCurrentWidget(self.ui.pageUsername)
            self.ui.btnBack.setVisible(False)
            self.ui.btnNext.setVisible(True)
            self.ui.btnNext.setEnabled(True if self.ui.txtUser.text() else False)
            self.ui.btnOk.setVisible(False)
        elif self.ui.stk.currentWidget() == self.ui.pageRepo:
            username = self.ui.txtUser.text()
            self.ui.lblPromptRepo.setTextFormat(Qt.RichText)
            self.ui.lblPromptRepo.setText("Repositories for <b>{}</b>".format(htmlescape(username)))
            try:
                repolist = self._get_repolist_for_user(username)
                if self.ui.cbxRepos.repolist != repolist:
                    with BlockedSignals(self.ui.cbxRepos):
                        self.ui.cbxRepos.clear()
                        for repo in repolist:
                            self.ui.cbxRepos.addItem(repo)
                        self.ui.cbxRepos.repolist = repolist
            except Exception as e:
                logger.debug("Ignoring exception ... %r", e)
                logger.exception("Ignoring exception")
                pass
            self.ui.btnBack.setVisible(True)
            self.ui.btnNext.setVisible(False)
            self.ui.btnOk.setVisible(True)
            self.ui.btnOk.setEnabled(True if self.ui.cbxRepos.currentText() else False)
            

    def _get_repolist_for_user(self, username):

        # use auth_token, if set
        from . import bibolamaziapp
        auth_token = bibolamaziapp.app_filterpackage_providers['github'].getAuthToken()

        if auth_token:
            G = github.Github(auth_token)
        else:
            G = github.Github()

        repo_list = []
        for repo in G.get_user(username).get_repos():
            repo_list.append(repo.name)

        logger.debug("Got repository list for user %r -> %r", username, repo_list)

        return repo_list

            
